"""Command line user interface to create force sets files."""

# Copyright (C) 2024 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import copy
import sys
from typing import Optional

import numpy as np
from phonopy.cui.create_force_sets import (
    check_agreement_of_supercell_positions,
    check_agreements_of_displacements,
    check_number_of_force_files,
)
from phonopy.cui.load_helper import get_nac_params
from phonopy.cui.phonopy_script import file_exists, files_exist, print_error
from phonopy.file_IO import is_file_phonopy_yaml, parse_FORCE_SETS, write_FORCE_SETS
from phonopy.interface.calculator import get_calc_dataset
from phonopy.structure.atoms import PhonopyAtoms

from phono3py.cui.settings import Phono3pySettings
from phono3py.file_IO import (
    get_length_of_first_line,
    parse_FORCES_FC2,
    write_FORCES_FC2,
    write_FORCES_FC3,
)
from phono3py.interface.phono3py_yaml import (
    Phono3pyYaml,
    displacements_yaml_lines_type1,
)
from phono3py.interface.wien2k import get_fc3_calc_dataset_wien2k


def create_FORCES_FC3_and_FORCES_FC2(
    settings: Phono3pySettings,
    cell_filename: Optional[str],
    log_level: int = 0,
):
    """Create FORCES_FC3 and FORCES_FC2 or phono3py_params.yaml from files.

    With settings.save_params=True, phono3py_params.yaml is created instead of
    FORCES_FC3 and FORCES_FC2. To create phono3py_params.yaml, at least fc3
    forces have to be collected.

    """
    interface_mode = settings.calculator

    disp_filename_candidates = [
        "phono3py_disp.yaml",
    ]

    if log_level:
        print("")

    if (
        log_level
        and cell_filename is not None
        and not is_file_phonopy_yaml(cell_filename, keyword="phono3py")
    ):
        print(f'*Unnecessary to specify "{cell_filename}".')

    if cell_filename is not None and is_file_phonopy_yaml(
        cell_filename, keyword="phono3py"
    ):
        disp_filename_candidates.insert(0, cell_filename)
    disp_filenames = files_exist(
        disp_filename_candidates, log_level=log_level, is_any=True
    )
    disp_filename = disp_filenames[0]
    ph3py_yaml = Phono3pyYaml(settings={"force_sets": True})
    ph3py_yaml.read(disp_filename)
    if ph3py_yaml.calculator is not None:
        interface_mode = ph3py_yaml.calculator  # overwrite

    if interface_mode is not None:
        if log_level:
            print(f"Calculator interface: {interface_mode}")

    if settings.create_forces_fc3 or settings.create_forces_fc3_file:
        calc_dataset_fc3 = _get_force_sets_fc3(
            settings,
            ph3py_yaml.dataset,
            ph3py_yaml.supercell,
            disp_filename,
            interface_mode,
            log_level,
        )
        if not calc_dataset_fc3["forces"]:
            if log_level:
                print('"FORCES_FC3" could not be created.')
                print_error()
            sys.exit(1)

    if settings.create_forces_fc2:
        calc_dataset_fc2 = _get_force_sets_fc2(
            settings,
            ph3py_yaml.phonon_dataset,
            ph3py_yaml.phonon_supercell,
            disp_filename,
            interface_mode,
            log_level,
        )
        if not calc_dataset_fc2["forces"]:
            if log_level:
                print('"FORCES_FC2" could not be created.')
                print_error()
            sys.exit(1)
    else:
        calc_dataset_fc2 = None

    if settings.save_params:
        fc3_yaml_filename = "phono3py_params.yaml"
        if not (settings.create_forces_fc3 or settings.create_forces_fc3_file):
            if log_level:
                print(f'When creating "{fc3_yaml_filename}" with force sets for fc2, ')
                print("force sets for fc3 have to be collected simultaneously.")
                print(f'"{fc3_yaml_filename}" could not be created.')
                print("")
                print_error()
            sys.exit(1)

        _set_forces_and_nac_params(
            ph3py_yaml, settings, calc_dataset_fc3, calc_dataset_fc2
        )

        with open(fc3_yaml_filename, "w") as w:
            w.write(str(ph3py_yaml))
            if log_level:
                print(f'"{fc3_yaml_filename}" has been created.')
    else:
        if settings.create_forces_fc3 or settings.create_forces_fc3_file:
            write_FORCES_FC3(
                ph3py_yaml.dataset,
                forces_fc3=calc_dataset_fc3["forces"],
                filename="FORCES_FC3",
            )
            if log_level:
                print('"FORCES_FC3" has been created.')

        if settings.create_forces_fc2:
            write_FORCES_FC2(
                ph3py_yaml.phonon_dataset,
                forces_fc2=calc_dataset_fc2["forces"],
                filename="FORCES_FC2",
            )
            if log_level:
                print('"FORCES_FC2" has been created.')


def create_FORCES_FC2_from_FORCE_SETS(log_level):
    """Convert FORCE_SETS to FORCES_FC2."""
    filename = "FORCE_SETS"
    file_exists(filename, log_level=log_level)
    disp_dataset = parse_FORCE_SETS(filename=filename)
    write_FORCES_FC2(disp_dataset)

    if log_level:
        print("")
        print("FORCES_FC2 has been created from FORCE_SETS.")
        print("The following yaml lines should replace respective part of")
        print("phono3py_disp.yaml made with --dim-fc2=dim_of_FORCE_SETS.")

        print("")
        print("\n".join(displacements_yaml_lines_type1(disp_dataset)))


def create_FORCE_SETS_from_FORCES_FCx(
    phonon_smat, cell_filename: Optional[str], log_level
):
    """Convert FORCES_FC3 or FORCES_FC2 to FORCE_SETS."""
    if cell_filename is not None and is_file_phonopy_yaml(
        cell_filename, keyword="phono3py"
    ):
        disp_filename = cell_filename
    else:
        disp_filename = "phono3py_disp.yaml"
    if phonon_smat is not None:
        forces_filename = "FORCES_FC2"
    else:
        forces_filename = "FORCES_FC3"

    if log_level:
        print(f'Displacement dataset is read from "{disp_filename}".')
        print(f'Forces are read from "{forces_filename}"')

    with open(forces_filename, "r") as f:
        len_first_line = get_length_of_first_line(f)

    if len_first_line == 3:
        file_exists(disp_filename, log_level=log_level)
        file_exists(forces_filename, log_level=log_level)
        ph3yml = Phono3pyYaml()
        ph3yml.read(disp_filename)
        if phonon_smat is None:
            dataset = copy.deepcopy(ph3yml.dataset)
            smat = ph3yml.supercell_matrix
        else:
            dataset = copy.deepcopy(ph3yml.phonon_dataset)
            smat = ph3yml.phonon_supercell_matrix

        if smat is None or (phonon_smat is not None and (phonon_smat != smat).any()):
            if log_level:
                print("")
                print("Supercell matrix is inconsistent.")
                print(f'Supercell matrix read from "{disp_filename}":')
                print(smat)
                print("Supercell matrix given by --dim-fc2:")
                print(phonon_smat)
                print_error()
            sys.exit(1)

        parse_FORCES_FC2(dataset, filename=forces_filename)
        write_FORCE_SETS(dataset)

        if log_level:
            print("FORCE_SETS has been created.")
    else:
        if log_level:
            print(
                "The file format of %s is already readable by phonopy."
                % forces_filename
            )


def _get_force_sets_fc2(
    settings: Phono3pySettings,
    disp_dataset: dict,
    supercell: PhonopyAtoms,
    disp_filename: str,
    interface_mode: Optional[str],
    log_level: int,
) -> dict:
    interface_mode = settings.calculator
    if log_level:
        print(f'FC2 displacement dataset was read from "{disp_filename}".')

    if "first_atoms" in disp_dataset:  # type-1
        num_atoms = disp_dataset["natom"]
        num_disps = len(disp_dataset["first_atoms"])
    elif "displacements" in disp_dataset:  # type-2:
        num_disps = len(disp_dataset["displacements"])
        num_atoms = len(disp_dataset["displacements"][0])
    else:
        raise RuntimeError("FC2 displacement dataset is broken.")

    force_filenames = settings.create_forces_fc2
    for filename in force_filenames:
        file_exists(filename, log_level=log_level)

    if log_level > 0:
        print(f"  Number of displacements: {num_disps}")
        print(f"  Number of supercell files: {len(force_filenames)}")

    calc_dataset = get_calc_dataset(
        interface_mode,
        num_atoms,
        force_filenames,
        verbose=(log_level > 0),
    )
    force_sets = calc_dataset["forces"]
    if "points" in calc_dataset:
        if filename := check_agreements_of_displacements(
            supercell, disp_dataset, calc_dataset["points"], force_filenames
        ):
            raise RuntimeError(
                f'Displacements don\'t match with atomic positions in "{filename}".'
            )

    if settings.subtract_forces or settings.subtract_forces_fc2:
        if settings.subtract_forces_fc2:
            force_filename = settings.subtract_forces_fc2
        else:
            force_filename = settings.subtract_forces
        file_exists(force_filename, log_level=log_level)
        calc_dataset_zero = get_calc_dataset(
            interface_mode,
            num_atoms,
            [
                force_filename,
            ],
            verbose=(log_level > 0),
        )
        if "points" in calc_dataset_zero:
            if check_agreement_of_supercell_positions(
                supercell, calc_dataset_zero["points"][0]
            ):
                raise RuntimeError(
                    "Supercell doesn't match with atomic positions in "
                    f'"{force_filename}".'
                )
        force_set_zero = calc_dataset_zero["forces"][0]
        for fs in force_sets:
            fs -= force_set_zero

        if log_level > 0:
            print(
                f'Forces in "{force_filename}" were subtracted from supercell forces.'
            )

    if log_level > 0 and "first_atoms" not in disp_dataset:
        if len(force_filenames) < num_disps:
            print("** Number of supercell files is less than displacements. **")

    if log_level > 0:
        print("")

    return calc_dataset


def _get_force_sets_fc3(
    settings: Phono3pySettings,
    disp_dataset: dict,
    supercell: PhonopyAtoms,
    disp_filename: str,
    interface_mode: Optional[str],
    log_level: int,
) -> dict:
    if log_level:
        print(f'FC3 Displacement dataset was read from "{disp_filename}".')

    if "first_atoms" in disp_dataset:  # type-1
        num_atoms = disp_dataset["natom"]
        num_disps = len(disp_dataset["first_atoms"])
        for d1 in disp_dataset["first_atoms"]:
            for d2 in d1["second_atoms"]:
                if "included" not in d2 or d2["included"]:
                    num_disps += 1
    elif "displacements" in disp_dataset:  # type-2:
        num_disps = len(disp_dataset["displacements"])
        num_atoms = len(disp_dataset["displacements"][0])
    else:
        raise RuntimeError("FC3 displacement dataset is broken.")

    if settings.create_forces_fc3_file:
        file_exists(settings.create_forces_fc3_file, log_level=log_level)
        force_filenames = [x.strip() for x in open(settings.create_forces_fc3_file)]
    else:
        force_filenames = settings.create_forces_fc3

    for filename in force_filenames:
        file_exists(filename, log_level=log_level)

    if log_level > 0:
        print(f"  Number of supercell files: {len(force_filenames)}")
        print(f'  Number of displacements in "{disp_filename}": {num_disps}')

    if "first_atoms" in disp_dataset and not check_number_of_force_files(
        num_disps, force_filenames, disp_filename
    ):  # type-1
        calc_dataset = {"forces": []}
    else:  # type-2
        if interface_mode == "wien2k":
            calc_dataset = get_fc3_calc_dataset_wien2k(
                force_filenames,
                supercell,
                disp_dataset,
                #                wien2k_P1_mode=wien2k_P1_mode,
                #                symmetry_tolerance=symmetry_tolerance,
                verbose=(log_level > 0),
            )
        else:
            calc_dataset = get_calc_dataset(
                interface_mode,
                num_atoms,
                force_filenames,
                verbose=(log_level > 0),
            )
        force_sets = calc_dataset["forces"]
        if "points" in calc_dataset:
            if filename := check_agreements_of_displacements(
                supercell, disp_dataset, calc_dataset["points"], force_filenames
            ):
                raise RuntimeError(
                    f'Displacements don\'t match with atomic positions in "{filename}".'
                )

    if settings.subtract_forces:
        force_filename = settings.subtract_forces
        file_exists(force_filename, log_level=log_level)
        calc_dataset_zero = get_calc_dataset(
            interface_mode,
            num_atoms,
            [
                force_filename,
            ],
            verbose=(log_level > 0),
        )
        force_set_zero = calc_dataset_zero["forces"][0]
        if "points" in calc_dataset_zero:
            if check_agreement_of_supercell_positions(
                supercell, calc_dataset_zero["points"][0]
            ):
                raise RuntimeError(
                    "Supercell doesn't match with atomic positions in "
                    f'"{force_filename}".'
                )

        for fs in force_sets:
            fs -= force_set_zero

        if log_level > 0:
            print(
                f"Forces in '{force_filename}' were subtracted from supercell forces."
            )

    if log_level > 0 and "first_atoms" not in disp_dataset:
        if len(force_filenames) < num_disps:
            print("** Number of supercell files is less than displacements. **")

    if log_level > 0:
        print("")

    return calc_dataset


def _set_forces_and_nac_params(
    ph3py_yaml: Phono3pyYaml,
    settings,
    calc_dataset_fc3: dict,
    calc_dataset_fc2: Optional[dict],
):
    """Set forces, energies and nac_params to phono3py_yaml."""
    if "first_atoms" in ph3py_yaml.dataset:
        count = len(ph3py_yaml.dataset["first_atoms"])
        for i, d1 in enumerate(ph3py_yaml.dataset["first_atoms"]):
            d1["forces"] = calc_dataset_fc3["forces"][i]
            if "supercell_energies" in calc_dataset_fc3:
                d1["supercell_energy"] = float(
                    calc_dataset_fc3["supercell_energies"][i]
                )
            for d2 in d1["second_atoms"]:
                if "included" not in d2 or d2["included"]:
                    d2["forces"] = calc_dataset_fc3["forces"][count]
                    if "supercell_energies" in calc_dataset_fc3:
                        d2["supercell_energy"] = float(
                            calc_dataset_fc3["supercell_energies"][count]
                        )
                    count += 1
    else:
        ph3py_yaml.dataset["forces"] = np.array(
            calc_dataset_fc3["forces"], dtype="double", order="C"
        )
        if "supercell_energies" in calc_dataset_fc3:
            ph3py_yaml.dataset["supercell_energies"] = np.array(
                calc_dataset_fc3["supercell_energies"], dtype="double"
            )
        if len(ph3py_yaml.dataset["forces"]) != len(
            ph3py_yaml.dataset["displacements"]
        ):
            ph3py_yaml.dataset["displacements"] = ph3py_yaml.dataset["displacements"][
                : len(ph3py_yaml.dataset["forces"])
            ]

    if settings.create_forces_fc2 and calc_dataset_fc2:
        if "first_atoms" in ph3py_yaml.phonon_dataset:
            for i, d in enumerate(ph3py_yaml.phonon_dataset["first_atoms"]):
                d["forces"] = calc_dataset_fc2["forces"][i]
                if "supercell_energies" in calc_dataset_fc2:
                    d["supercell_energy"] = float(
                        calc_dataset_fc2["supercell_energies"][i]
                    )
        else:
            ph3py_yaml.phonon_dataset["forces"] = np.array(
                calc_dataset_fc2["forces"], dtype="double", order="C"
            )
            if "supercell_energies" in calc_dataset_fc2:
                ph3py_yaml.phonon_dataset["supercell_energies"] = np.array(
                    calc_dataset_fc2["supercell_energies"], dtype="double"
                )
            if len(ph3py_yaml.phonon_dataset["forces"]) != len(
                ph3py_yaml.phonon_dataset["displacements"]
            ):
                ph3py_yaml.phonon_dataset["displacements"] = ph3py_yaml.phonon_dataset[
                    "displacements"
                ][: len(ph3py_yaml.phonon_dataset["forces"])]

    if ph3py_yaml.nac_params is None:
        nac_params = get_nac_params(primitive=ph3py_yaml.primitive)
        if nac_params:
            ph3py_yaml.nac_params = nac_params



"""Tests of Phono3py API."""

from __future__ import annotations

from pathlib import Path

import numpy as np
import pytest
from phonopy.harmonic.force_constants import get_drift_force_constants
from phonopy.interface.pypolymlp import PypolymlpParams

from phono3py import Phono3py
from phono3py.phonon3.fc3 import get_drift_fc3

cwd = Path(__file__).parent


def test_displacements_setter_NaCl(nacl_pbe: Phono3py):
    """Test Phono3py.displacements setter.

    Just check no error.

    """
    ph3_in = nacl_pbe
    displacements = ph3_in.displacements
    ph3 = Phono3py(
        ph3_in.unitcell,
        supercell_matrix=ph3_in.supercell_matrix,
        primitive_matrix=ph3_in.primitive_matrix,
    )
    ph3.displacements = displacements


def test_displacements_setter_Si(si_pbesol_111_222_fd: Phono3py):
    """Test Phono3py.displacements setter and Phono3py.phonon_displacements setter.

    Just check no error.

    """
    ph3_in = si_pbesol_111_222_fd
    displacements = ph3_in.displacements
    phonon_displacements = ph3_in.phonon_displacements
    ph3 = Phono3py(
        ph3_in.unitcell,
        supercell_matrix=ph3_in.supercell_matrix,
        phonon_supercell_matrix=ph3_in.phonon_supercell_matrix,
        primitive_matrix=ph3_in.primitive_matrix,
    )
    ph3.displacements = displacements
    ph3.phonon_displacements = phonon_displacements


def test_type1_forces_energies_setter_Si(si_111_222_fd: Phono3py):
    """Test type1 supercell_energies, phonon_supercell_energies attributes."""
    ph3_in = si_111_222_fd
    ref_ph_supercell_energies = [-346.85204143]
    ref_supercell_energies = [
        -43.3509760,
        -43.33608775,
        -43.35352904,
        -43.34370672,
        -43.34590849,
        -43.34540162,
        -43.34421408,
        -43.34481089,
        -43.34703607,
        -43.34241924,
        -43.34786243,
        -43.34168203,
        -43.34274245,
        -43.34703607,
        -43.34786243,
        -43.34184454,
    ]
    np.testing.assert_allclose(ph3_in.supercell_energies, ref_supercell_energies)
    np.testing.assert_allclose(
        ph3_in.phonon_supercell_energies, ref_ph_supercell_energies
    )

    ref_force00 = [-0.4109520800000000, 0.0000000100000000, 0.0000000300000000]
    ref_force_last = [0.1521426300000000, 0.0715600600000000, -0.0715600700000000]
    ref_ph_force00 = [-0.4027479600000000, 0.0000000200000000, 0.0000001000000000]
    np.testing.assert_allclose(ph3_in.forces[0, 0], ref_force00)
    np.testing.assert_allclose(ph3_in.forces[-1, -1], ref_force_last)
    np.testing.assert_allclose(ph3_in.phonon_forces[0, 0], ref_ph_force00)

    ph3 = Phono3py(
        ph3_in.unitcell,
        supercell_matrix=ph3_in.supercell_matrix,
        phonon_supercell_matrix=ph3_in.phonon_supercell_matrix,
        primitive_matrix=ph3_in.primitive_matrix,
    )
    ph3.dataset = ph3_in.dataset
    ph3.phonon_dataset = ph3_in.phonon_dataset

    ph3.supercell_energies = ph3_in.supercell_energies + 1
    ph3.phonon_supercell_energies = ph3_in.phonon_supercell_energies + 1
    np.testing.assert_allclose(ph3_in.supercell_energies + 1, ph3.supercell_energies)
    np.testing.assert_allclose(
        ph3_in.phonon_supercell_energies + 1, ph3.phonon_supercell_energies
    )

    ph3.forces = ph3_in.forces + 1
    ph3.phonon_forces = ph3_in.phonon_forces + 1
    np.testing.assert_allclose(ph3_in.forces + 1, ph3.forces)
    np.testing.assert_allclose(ph3_in.phonon_forces + 1, ph3.phonon_forces)


def test_type2_forces_energies_setter_Si(si_111_222_rd: Phono3py):
    """Test type2 supercell_energies, phonon_supercell_energies attributes."""
    ph3_in = si_111_222_rd
    ref_ph_supercell_energies = [
        -346.81061270,  # 1
        -346.81263617,  # 2
    ]
    ref_supercell_energies = [
        -43.35270268,  # 1
        -43.35211687,  # 2
        -43.35122776,  # 3
        -43.35226673,  # 4
        -43.35146358,  # 5
        -43.35133209,  # 6
        -43.35042212,  # 7
        -43.35008442,  # 8
        -43.34968796,  # 9
        -43.35348999,  # 10
        -43.35134937,  # 11
        -43.35335251,  # 12
        -43.35160892,  # 13
        -43.35009115,  # 14
        -43.35202797,  # 15
        -43.35076370,  # 16
        -43.35174477,  # 17
        -43.35107001,  # 18
        -43.35037949,  # 19
        -43.35126123,  # 20
    ]
    np.testing.assert_allclose(ph3_in.supercell_energies, ref_supercell_energies)
    np.testing.assert_allclose(
        ph3_in.phonon_supercell_energies, ref_ph_supercell_energies
    )

    ref_force00 = [0.0445647800000000, 0.1702929900000000, 0.0913398200000000]
    ref_force_last = [-0.1749668700000000, 0.0146997300000000, -0.1336066300000000]
    ref_ph_force00 = [-0.0161598900000000, -0.1161657500000000, 0.1399128100000000]
    ref_ph_force_last = [0.1049486700000000, 0.0795870900000000, 0.1062164600000000]

    np.testing.assert_allclose(ph3_in.forces[0, 0], ref_force00)
    np.testing.assert_allclose(ph3_in.forces[-1, -1], ref_force_last)
    np.testing.assert_allclose(ph3_in.phonon_forces[0, 0], ref_ph_force00)
    np.testing.assert_allclose(ph3_in.phonon_forces[-1, -1], ref_ph_force_last)

    ph3 = Phono3py(
        ph3_in.unitcell,
        supercell_matrix=ph3_in.supercell_matrix,
        phonon_supercell_matrix=ph3_in.phonon_supercell_matrix,
        primitive_matrix=ph3_in.primitive_matrix,
    )
    ph3.dataset = ph3_in.dataset
    ph3.phonon_dataset = ph3_in.phonon_dataset
    ph3.supercell_energies = ph3_in.supercell_energies + 1
    ph3.phonon_supercell_energies = ph3_in.phonon_supercell_energies + 1

    np.testing.assert_allclose(ph3_in.supercell_energies + 1, ph3.supercell_energies)
    np.testing.assert_allclose(
        ph3_in.phonon_supercell_energies + 1, ph3.phonon_supercell_energies
    )

    ph3.forces = ph3_in.forces + 1
    ph3.phonon_forces = ph3_in.phonon_forces + 1
    np.testing.assert_allclose(ph3_in.forces + 1, ph3.forces)
    np.testing.assert_allclose(ph3_in.phonon_forces + 1, ph3.phonon_forces)


def test_use_pypolymlp_mgo(mgo_222rd_444rd: Phono3py):
    """Test use_pypolymlp in produce_fc3."""
    pytest.importorskip("pypolymlp")
    pytest.importorskip("symfc")

    ph3_in = mgo_222rd_444rd
    ph3 = Phono3py(
        ph3_in.unitcell,
        supercell_matrix=ph3_in.supercell_matrix,
        phonon_supercell_matrix=ph3_in.phonon_supercell_matrix,
        primitive_matrix=ph3_in.primitive_matrix,
        log_level=2,
    )
    ph3.mlp_dataset = {
        "displacements": ph3_in.displacements[:10],
        "forces": ph3_in.forces[:10],
        "supercell_energies": ph3_in.supercell_energies[:10],
    }
    ph3.phonon_dataset = ph3_in.phonon_dataset
    ph3.nac_params = ph3_in.nac_params
    ph3.displacements = ph3_in.displacements[:100]

    atom_energies = {"Mg": -0.00896717, "O": -0.95743902}
    params = PypolymlpParams(gtinv_maxl=(4, 4), atom_energies=atom_energies)

    ph3.generate_displacements(distance=0.001, is_plusminus=True)
    ph3.develop_mlp(params=params)
    ph3.evaluate_mlp()
    ph3.produce_fc3(fc_calculator="symfc")
    ph3.produce_fc2(fc_calculator="symfc")

    ph3.mesh_numbers = 30
    ph3.init_phph_interaction()
    ph3.run_thermal_conductivity(temperatures=[300])
    assert (
        pytest.approx(63.0018546, abs=1e-1) == ph3.thermal_conductivity.kappa[0, 0, 0]
    )


@pytest.mark.parametrize("is_compact_fc", [True, False])
def test_symmetrize_fc_traditional(si_pbesol: Phono3py, is_compact_fc: bool):
    """Test symmetrize_fc3 and symmetrize_fc2 with traditional approach."""
    ph3 = Phono3py(
        si_pbesol.unitcell,
        supercell_matrix=si_pbesol.supercell_matrix,
        primitive_matrix=si_pbesol.primitive_matrix,
        log_level=2,
    )
    ph3.dataset = si_pbesol.dataset
    ph3.produce_fc3(is_compact_fc=is_compact_fc)
    assert ph3.fc3 is not None
    assert ph3.fc2 is not None

    v1, v2, v3, _, _, _ = get_drift_fc3(ph3.fc3, primitive=ph3.primitive)
    np.testing.assert_allclose(
        np.abs([v1, v2, v3]), [1.755065e-01, 1.749287e-01, 3.333333e-05], atol=1e-6
    )
    ph3.symmetrize_fc3(options="level=3")
    v1_sym, v2_sym, v3_sym, _, _, _ = get_drift_fc3(ph3.fc3, primitive=ph3.primitive)
    if is_compact_fc:
        np.testing.assert_allclose(
            np.abs([v1_sym, v2_sym, v3_sym]), 1.217081e-05, atol=1e-6
        )
    else:
        np.testing.assert_allclose(
            np.abs([v1_sym, v2_sym, v3_sym]), 1.421085e-14, atol=1e-6
        )

    v1_sym, v2_sym, _, _ = get_drift_force_constants(ph3.fc2, primitive=ph3.primitive)
    if is_compact_fc:
        np.testing.assert_allclose(np.abs([v1_sym, v2_sym]), 1.0e-06, atol=1e-6)
    else:
        np.testing.assert_allclose(np.abs([v1_sym, v2_sym]), 1.0e-06, atol=1e-6)


@pytest.mark.parametrize("is_compact_fc", [True, False])
def test_symmetrize_fc_symfc(si_pbesol: Phono3py, is_compact_fc: bool):
    """Test symmetrize_fc3 and symmetrize_fc2 with symfc."""
    pytest.importorskip("symfc")

    ph3 = Phono3py(
        si_pbesol.unitcell,
        supercell_matrix=si_pbesol.supercell_matrix,
        primitive_matrix=si_pbesol.primitive_matrix,
        log_level=2,
    )
    ph3.dataset = si_pbesol.dataset
    ph3.produce_fc3(is_compact_fc=is_compact_fc)
    assert ph3.fc3 is not None
    assert ph3.fc2 is not None

    ph3.symmetrize_fc3(use_symfc_projector=True)
    v1_sym, v2_sym, v3_sym, _, _, _ = get_drift_fc3(ph3.fc3, primitive=ph3.primitive)
    np.testing.assert_allclose([v1_sym, v2_sym, v3_sym], 0, atol=1e-6)
    ph3.symmetrize_fc2(use_symfc_projector=True)
    v1_sym, v2_sym, _, _ = get_drift_force_constants(ph3.fc2, primitive=ph3.primitive)
    np.testing.assert_allclose([v1_sym, v2_sym], 0, atol=1e-6)



"""Tests of displacements.py."""

import itertools
from typing import Literal, Optional, Union

import numpy as np
import pytest
from phonopy.structure.atoms import PhonopyAtoms

import phono3py
from phono3py import Phono3py
from phono3py.phonon3.displacement_fc3 import get_smallest_vector_of_atom_pair

distances_NaCl = [
    0.0000000,
    5.6903015,
    5.6903015,
    8.0473015,
    5.6903015,
    8.0473015,
    8.0473015,
    9.8558913,
    4.0236508,
    6.9691675,
    4.0236508,
    6.9691675,
    4.0236508,
    6.9691675,
    4.0236508,
    6.9691675,
    4.0236508,
    4.0236508,
    6.9691675,
    6.9691675,
    4.0236508,
    4.0236508,
    6.9691675,
    6.9691675,
    4.0236508,
    4.0236508,
    4.0236508,
    4.0236508,
    6.9691675,
    6.9691675,
    6.9691675,
    6.9691675,
    4.9279456,
    4.9279456,
    4.9279456,
    4.9279456,
    4.9279456,
    4.9279456,
    4.9279456,
    4.9279456,
    2.8451507,
    2.8451507,
    6.3619505,
    6.3619505,
    6.3619505,
    6.3619505,
    8.5354522,
    8.5354522,
    2.8451507,
    6.3619505,
    2.8451507,
    6.3619505,
    6.3619505,
    8.5354522,
    6.3619505,
    8.5354522,
    2.8451507,
    6.3619505,
    6.3619505,
    8.5354522,
    2.8451507,
    6.3619505,
    6.3619505,
    8.5354522,
]


@pytest.mark.parametrize(
    "is_plusminus,distance,number_of_snapshots",
    itertools.product([False, True], [None, 0.06], [8, "auto"]),
)
def test_random_disps_agno2(
    agno2_cell: PhonopyAtoms,
    is_plusminus: bool,
    distance: Optional[float],
    number_of_snapshots: Union[int, Literal["auto"]],
):
    """Test random displacements."""
    pytest.importorskip("symfc")

    ph3 = phono3py.load(unitcell=agno2_cell, supercell_matrix=[2, 1, 2])
    ph3.generate_displacements(
        number_of_snapshots=number_of_snapshots,
        distance=distance,
        is_plusminus=is_plusminus,
        number_estimation_factor=2,
    )

    ph3.generate_fc2_displacements(
        number_of_snapshots=number_of_snapshots,
        distance=distance,
        is_plusminus=is_plusminus,
    )

    for d, n_d in zip((ph3.displacements, ph3.phonon_displacements), (96, 4)):
        if number_of_snapshots == "auto":
            assert len(d) == n_d * (is_plusminus + 1)
        else:
            assert len(d) == 8 * (is_plusminus + 1)

        if distance is None:
            np.testing.assert_allclose(
                np.linalg.norm(d, axis=2).ravel(), 0.03, atol=1e-8
            )
        else:
            np.testing.assert_allclose(
                np.linalg.norm(d, axis=2).ravel(), 0.06, atol=1e-8
            )
        if is_plusminus:
            np.testing.assert_allclose(d[: len(d) // 2], -d[len(d) // 2 :], atol=1e-8)


def test_duplicates_agno2(agno2_cell: PhonopyAtoms):
    """Test duplicated pairs of displacements."""
    ph3 = phono3py.load(unitcell=agno2_cell, supercell_matrix=[1, 1, 1])
    ph3.generate_displacements()
    duplicates_ref = [
        [106, 22],
        [220, 80],
        [252, 81],
        [221, 96],
        [253, 97],
        [290, 142],
        [348, 244],
        [364, 245],
        [349, 276],
        [365, 277],
        [119, 0],
        [261, 0],
        [229, 0],
        [260, 0],
        [228, 0],
    ]
    np.testing.assert_equal(duplicates_ref, ph3.dataset["duplicates"])


def test_nacl_pbe(nacl_pbe: Phono3py):
    """Test generated displacements and duplicates."""
    ph3 = Phono3py(
        nacl_pbe.unitcell,
        supercell_matrix=nacl_pbe.supercell_matrix,
        primitive_matrix=nacl_pbe.primitive_matrix,
    )
    ph3.generate_displacements()
    duplicates_ref = [[77, 41]]
    ph3.dataset["duplicates"]
    np.testing.assert_equal(duplicates_ref, ph3.dataset["duplicates"])

    pairs_ref = [
        0,
        0,
        0,
        1,
        0,
        2,
        0,
        3,
        0,
        6,
        0,
        7,
        0,
        8,
        0,
        9,
        0,
        16,
        0,
        17,
        0,
        18,
        0,
        19,
        0,
        32,
        0,
        33,
        0,
        40,
        0,
        41,
        0,
        42,
        0,
        43,
        0,
        46,
        0,
        47,
        0,
        48,
        0,
        49,
        0,
        52,
        0,
        53,
        32,
        0,
        32,
        1,
        32,
        8,
        32,
        9,
        32,
        10,
        32,
        11,
        32,
        14,
        32,
        15,
        32,
        16,
        32,
        17,
        32,
        20,
        32,
        21,
        32,
        32,
        32,
        33,
        32,
        34,
        32,
        35,
        32,
        38,
        32,
        39,
        32,
        40,
        32,
        41,
        32,
        48,
        32,
        49,
        32,
        50,
        32,
        51,
    ]
    pairs = []
    for first_atoms in ph3.dataset["first_atoms"]:
        n1 = first_atoms["number"]
        n2s = np.unique(
            [second_atoms["number"] for second_atoms in first_atoms["second_atoms"]]
        )
        pairs += [[n1, n2] for n2 in n2s]
    # print("".join(["%d, " % i for i in np.array(pairs).ravel()]))

    np.testing.assert_equal(pairs_ref, np.array(pairs).ravel())


def test_get_smallest_vector_of_atom_pair(nacl_pbe):
    """Test get_smallest_vector_of_atom_pair."""
    ph3 = nacl_pbe
    distances = []
    for i in range(len(ph3.supercell)):
        vec = get_smallest_vector_of_atom_pair(i, 0, ph3.supercell, 1e-5)
        if vec.ndim == 2:
            vec = vec[0]
        distances.append(np.linalg.norm(np.dot(vec, ph3.supercell.cell)))

    # _show(distances)
    np.testing.assert_allclose(distances_NaCl, distances, rtol=0, atol=1e-6)


def _show(vals):
    for i, v in enumerate(vals):
        print("%.7f, " % v, end="")
        if (i + 1) % 5 == 0:
            print("")



"""Lattice thermal conductivity calculation with RTA."""

# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

from typing import Optional, Union

import numpy as np

from phono3py.conductivity.base import ConductivityComponents
from phono3py.conductivity.rta_base import ConductivityRTABase
from phono3py.phonon3.interaction import Interaction


class ConductivityRTA(ConductivityRTABase):
    """Lattice thermal conductivity calculation with RTA."""

    def __init__(
        self,
        interaction: Interaction,
        grid_points: Optional[np.ndarray] = None,
        temperatures: Optional[Union[list, np.ndarray]] = None,
        sigmas: Optional[Union[list, np.ndarray]] = None,
        sigma_cutoff: Optional[float] = None,
        is_isotope: bool = False,
        mass_variances: Optional[Union[list, np.ndarray]] = None,
        boundary_mfp: Optional[float] = None,  # in micrometer
        use_ave_pp: bool = False,
        is_kappa_star: bool = True,
        gv_delta_q: Optional[float] = None,
        is_full_pp: bool = False,
        read_pp: bool = False,
        store_pp: bool = False,
        pp_filename: Optional[float] = None,
        is_N_U: bool = False,
        is_gamma_detail: bool = False,
        is_frequency_shift_by_bubble: bool = False,
        log_level: int = 0,
    ):
        """Init method."""
        self._kappa = None
        self._mode_kappa = None

        super().__init__(
            interaction,
            grid_points=grid_points,
            temperatures=temperatures,
            sigmas=sigmas,
            sigma_cutoff=sigma_cutoff,
            is_isotope=is_isotope,
            mass_variances=mass_variances,
            boundary_mfp=boundary_mfp,
            use_ave_pp=use_ave_pp,
            is_kappa_star=is_kappa_star,
            is_full_pp=is_full_pp,
            read_pp=read_pp,
            store_pp=store_pp,
            pp_filename=pp_filename,
            is_N_U=is_N_U,
            is_gamma_detail=is_gamma_detail,
            is_frequency_shift_by_bubble=is_frequency_shift_by_bubble,
            log_level=log_level,
        )

        self._conductivity_components = ConductivityComponents(
            self._pp,
            self._grid_points,
            self._grid_weights,
            self._point_operations,
            self._rotations_cartesian,
            temperatures=self._temperatures,
            average_gv_over_kstar=self._average_gv_over_kstar,
            is_kappa_star=self._is_kappa_star,
            gv_delta_q=gv_delta_q,
            log_level=self._log_level,
        )

    @property
    def kappa(self):
        """Return kappa."""
        return self._kappa

    @property
    def mode_kappa(self):
        """Return mode_kappa."""
        return self._mode_kappa

    @property
    def gv_by_gv(self):
        """Return gv_by_gv at grid points where mode kappa are calculated."""
        return self._conductivity_components.gv_by_gv

    def _set_cv(self, i_gp, i_data):
        """Set cv for conductivity components."""
        self._conductivity_components.set_heat_capacities(i_gp, i_data)

    def _set_velocities(self, i_gp, i_data):
        """Set velocities for conductivity components."""
        self._conductivity_components.set_velocities(i_gp, i_data)

    def set_kappa_at_sigmas(self):
        """Calculate kappa from ph-ph interaction results."""
        if not self._pp.phonon_all_done:
            raise RuntimeError(
                "Phonon calculation has not been done yet. "
                "Run phono3py.run_phonon_solver() before this method."
            )
        if self._temperatures is None:
            raise RuntimeError(
                "Temperatures have not been set yet. "
                "Set temperatures before this method."
            )

        num_band = len(self._pp.primitive) * 3
        mode_heat_capacities = self._conductivity_components.mode_heat_capacities
        gv_by_gv = self._conductivity_components.gv_by_gv
        for i, _ in enumerate(self._grid_points):
            cv = mode_heat_capacities[:, i, :]
            gp = self._grid_points[i]
            frequencies = self._frequencies[gp]  # type: ignore

            # Kappa
            for j in range(len(self._sigmas)):
                for k in range(len(self._temperatures)):
                    g_sum = self._get_main_diagonal(i, j, k)
                    for ll in range(num_band):
                        if frequencies[ll] < self._pp.cutoff_frequency:
                            self._num_ignored_phonon_modes[j, k] += 1  # type: ignore
                            continue

                        old_settings = np.seterr(all="raise")
                        try:
                            self._mode_kappa[j, k, i, ll] = (  # type: ignore
                                gv_by_gv[i, ll]
                                * cv[k, ll]
                                / (g_sum[ll] * 2)
                                * self._conversion_factor
                            )
                        except FloatingPointError:
                            # supposed that g is almost 0 and |gv|=0
                            pass
                        except Exception:
                            print("=" * 26 + " Warning " + "=" * 26)
                            print(
                                " Unexpected physical condition of ph-ph "
                                "interaction calculation was found."
                            )
                            print(
                                " g=%f at gp=%d, band=%d, freq=%f"
                                % (g_sum[ll], gp, ll + 1, frequencies[ll])
                            )
                            print("=" * 61)
                        np.seterr(**old_settings)

        N = self.number_of_sampling_grid_points
        self._kappa = self._mode_kappa.sum(axis=2).sum(axis=2) / N

    def _allocate_values(self):
        if self._temperatures is None:
            raise RuntimeError(
                "Temperatures have not been set yet. "
                "Set temperatures before this method."
            )

        super()._allocate_values()

        num_band = len(self._pp.primitive) * 3
        num_grid_points = len(self._grid_points)
        num_temp = len(self._temperatures)

        # kappa* and mode_kappa* are accessed when all bands exist, i.e.,
        # num_band0==num_band.
        self._kappa = np.zeros(
            (len(self._sigmas), num_temp, 6), order="C", dtype="double"
        )
        self._mode_kappa = np.zeros(
            (len(self._sigmas), num_temp, num_grid_points, num_band, 6),
            order="C",
            dtype="double",
        )



import argparse
import os
import sys

import h5py
import numpy as np

epsilon = 1.0e-8


def get_options():
    """Parse command line options."""
    parser = argparse.ArgumentParser(description="Plot collision matrix eigenvalues")
    parser.add_argument(
        "--temperature",
        type=float,
        default=300.0,
        dest="temperature",
        help="Temperature to output data at",
    )
    parser.add_argument(
        "--savetxt",
        dest="savetxt",
        action="store_true",
        default=False,
        help="Write coleigs to text file",
    )
    parser.add_argument("filenames", nargs="*")
    args = parser.parse_args()
    return args


def get_t_index(temperatures, args):
    """Return temperature index."""
    if len(temperatures) > 29:
        t_index = 30
    else:
        t_index = 0
    for i, t in enumerate(temperatures):
        if np.abs(t - args.temperature) < epsilon:
            t_index = i
            break
    return t_index


def plot_one(ax, coleigs, temperatures, args):
    """Plot collision matrix eigenvalues at one temperature."""
    t_index = get_t_index(temperatures, args)
    coleigs_p = coleigs[t_index].copy()
    coleigs_n = coleigs[t_index].copy()
    coleigs_p[coleigs_p < 0] = np.nan
    coleigs_n[coleigs_n >= 0] = np.nan
    ax.semilogy(coleigs_p, "b.", markersize=1, clip_on=False)
    ax.semilogy(-coleigs_n, "r.", markersize=1, clip_on=False)
    ax.set_xlim(0, len(coleigs_p))


def plot_one_file(ax, args):
    """Plot collision matrix eigenvalues at one temperature from file."""
    filename = args.filenames[0]
    if os.path.isfile(filename):
        with h5py.File(filename, "r") as f:
            coleigs = f["collision_eigenvalues"][:]
            temperatures = f["temperature"][:]
        plot_one(ax, coleigs, temperatures, args)
    else:
        print("File %s doesn't exist." % filename)
        sys.exit(1)


def plot_more(ax, coleigs, temperatures, args):
    """Update plot by collision matrix eigenvalues from one file."""
    t_index = get_t_index(temperatures[0], args)
    y = [np.abs(v[t_index]) for v in coleigs]
    ax.semilogy(np.transpose(y), ".", markersize=5)
    ax.set_xlim(0, len(y[0]))


def plot_more_files(ax, args):
    """Plot collision matrix eigenvalues from multiple files."""
    temperatures = []
    coleigs = []
    for filename in args.filenames:
        if os.path.isfile(filename):
            with h5py.File(filename, "r") as f:
                coleigs.append(f["collision_eigenvalues"][:])
                temperatures.append(f["temperature"][:])
        else:
            print("File %s doesn't exist." % filename)
            sys.exit(1)
    plot_more(ax, coleigs, temperatures, args)


def plot(args: argparse.Namespace):
    """Plot collision matrix eigenvalues."""
    import matplotlib.pyplot as plt

    _, ax = plt.subplots()
    ax.xaxis.set_ticks_position("bottom")
    ax.yaxis.set_ticks_position("left")
    ax.xaxis.set_tick_params(which="both", direction="in")
    ax.yaxis.set_tick_params(which="both", direction="in")

    if len(args.filenames) == 1:
        plot_one_file(ax, args)
    else:
        plot_more_files(ax, args)

    # for filename in args.filenames:

    plt.show()


def write_dat(args: argparse.Namespace):
    """Write collision matrix eigenvalues to a file in gnuplot style."""
    with open("coleigs.dat", "w") as w:
        for filename in args.filenames:
            with h5py.File(filename) as f:
                print("#", file=w)
                print(f"# Filename {filename}", file=w)
                print("#", file=w)
                print("#             eigenvalue      log10(abs(eigenvalue))", file=w)
                log_vals = np.log10(np.abs(f["collision_eigenvalues"][:]))
                for coleigs, coleigs_log in zip(f["collision_eigenvalues"], log_vals):
                    for i, (val, val_log) in enumerate(zip(coleigs, coleigs_log)):
                        print(f"{i + 1:<10d} {val:15.8e}     {val_log:15.8f}", file=w)
                    print(file=w)
                print(file=w)
                print(file=w)


def main(args: argparse.Namespace):
    """Run phono3py-coleigplot."""
    if args.savetxt:
        write_dat(args)
    else:
        plot(args)


def run():
    """Run phono3py-coleigplot script."""
    main(get_options())



"""Utilities for lattice thermal conductivity calculation."""

# Copyright (C) 2022 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from typing import TYPE_CHECKING

from phono3py.file_IO import write_pp_to_hdf5
from phono3py.phonon3.interaction import Interaction
from phono3py.phonon3.triplets import get_all_triplets

if TYPE_CHECKING:
    from phono3py.conductivity.base import ConductivityBase


def select_colmat_solver(pinv_solver):
    """Return collision matrix solver id."""
    try:
        import phono3py._phono3py as phono3c

        default_solver = phono3c.default_colmat_solver()
    except ImportError:
        print("Phono3py C-routine is not compiled correctly.")
        default_solver = 4

    if not phono3c.include_lapacke():
        if pinv_solver in (1, 2, 6):
            raise RuntimeError(
                "Use pinv-solver 3, 4, or 5 because "
                "phono3py is not compiled with LAPACKE."
            )

    solver_numbers = (1, 2, 3, 4, 5, 6, 7)

    solver = pinv_solver
    if solver == 6:  # 6 must return 3 for not transposing unitary matrix.
        solver = 3
    if solver == 0:  # default solver
        if default_solver in (3, 4, 5):
            try:
                import scipy.linalg  # noqa F401
            except ImportError:
                solver = 1
            else:
                solver = default_solver
        else:
            solver = default_solver
    elif solver not in solver_numbers:
        solver = default_solver

    return solver


def write_pp_interaction(
    conductivity: "ConductivityBase",
    pp: Interaction,
    i,
    filename=None,
    compression="gzip",
):
    """Write ph-ph interaction strength in hdf5 file."""
    grid_point = conductivity.grid_points[i]
    sigmas = conductivity.sigmas
    sigma_cutoff = conductivity.sigma_cutoff_width
    mesh = conductivity.mesh_numbers
    triplets, weights, _, _ = pp.get_triplets_at_q()
    all_triplets = get_all_triplets(grid_point, pp.bz_grid)

    if len(sigmas) > 1:
        print("Multiple smearing parameters were given. The last one in ")
        print("ph-ph interaction calculations was written in the file.")

    write_pp_to_hdf5(
        mesh,
        pp=pp.interaction_strength,
        g_zero=pp.zero_value_positions,
        grid_point=grid_point,
        triplet=triplets,
        weight=weights,
        triplet_all=all_triplets,
        sigma=sigmas[-1],
        sigma_cutoff=sigma_cutoff,
        filename=filename,
        compression=compression,
    )



"""Tests of phono3py-load script."""

from __future__ import annotations

import os
import pathlib
import tempfile
from collections.abc import Sequence
from dataclasses import dataclass, fields

import h5py
import numpy as np
import pytest

import phono3py
from phono3py.cui.phono3py_script import main

cwd = pathlib.Path(__file__).parent


@dataclass
class MockArgs:
    """Mock args of ArgumentParser."""

    cell_filename: str | os.PathLike | None = None
    conf_filename: str | os.PathLike | None = None
    fc_calculator: str | None = None
    fc_calculator_options: str | None = None
    fc_symmetry: bool = True
    filename: Sequence[str | os.PathLike] | None = None
    force_sets_mode: bool = False
    force_sets_to_forces_fc2_mode: bool = False
    input_filename = None
    input_output_filename = None
    log_level: int | None = None
    is_bterta: bool | None = None
    is_lbte: bool | None = None
    is_wigner_kappa: bool | None = None
    mesh_numbers: Sequence | None = None
    mlp_params: str | None = None
    rd_number_estimation_factor: float | None = None
    output_filename = None
    output_yaml_filename: str | os.PathLike | None = None
    random_displacements: int | str | None = None
    show_num_triplets: bool = False
    temperatures: Sequence | None = None
    use_pypolymlp: bool = False
    write_grid_points: bool = False

    def __iter__(self):
        """Make self iterable to support in."""
        return (getattr(self, field.name) for field in fields(self))

    def __contains__(self, item):
        """Implement in operator."""
        return item in (field.name for field in fields(self))


def test_phono3py_load():
    """Test phono3py-load script."""
    with tempfile.TemporaryDirectory() as temp_dir:
        original_cwd = pathlib.Path.cwd()
        os.chdir(temp_dir)

        try:
            # Check sys.exit(0)
            argparse_control = _get_phono3py_load_args(
                cwd / ".." / "phono3py_params_Si-111-222.yaml",
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            argparse_control = _get_phono3py_load_args(
                "phono3py.yaml",
                is_bterta=True,
                temperatures=[
                    "300",
                ],
                mesh_numbers=["5", "5", "5"],
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            # Clean files created by phono3py-load script.
            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "kappa-m555.hdf5",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

        finally:
            os.chdir(original_cwd)


def test_phono3py_load_lbte():
    """Test phono3py-load script running direct solution."""
    with tempfile.TemporaryDirectory() as temp_dir:
        original_cwd = pathlib.Path.cwd()
        os.chdir(temp_dir)

        try:
            # Check sys.exit(0)
            argparse_control = _get_phono3py_load_args(
                cwd / ".." / "phono3py_params_Si-111-222.yaml",
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            argparse_control = _get_phono3py_load_args(
                "phono3py.yaml",
                is_lbte=True,
                temperatures=[
                    "300",
                ],
                mesh_numbers=["5", "5", "5"],
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            # Clean files created by phono3py-load script.
            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "kappa-m555.hdf5",
                "coleigs-m555.hdf5",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

        finally:
            os.chdir(original_cwd)


def test_phono3py_load_wigner_rta():
    """Test phono3py-load script running wigner rta."""
    with tempfile.TemporaryDirectory() as temp_dir:
        original_cwd = pathlib.Path.cwd()
        os.chdir(temp_dir)

        try:
            # Check sys.exit(0)
            argparse_control = _get_phono3py_load_args(
                cwd / ".." / "phono3py_params_Si-111-222.yaml",
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            argparse_control = _get_phono3py_load_args(
                "phono3py.yaml",
                is_bterta=True,
                is_wigner_kappa=True,
                temperatures=[
                    "300",
                ],
                mesh_numbers=["5", "5", "5"],
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            # for filename in pathlib.Path.cwd().iterdir():
            #     print(filename)

            # Clean files created by phono3py-load script.
            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "kappa-m555.hdf5",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

        finally:
            os.chdir(original_cwd)


def test_phono3py_load_wigner_lbte():
    """Test phono3py-load script running wigner rta."""
    with tempfile.TemporaryDirectory() as temp_dir:
        original_cwd = pathlib.Path.cwd()
        os.chdir(temp_dir)

        try:
            # Check sys.exit(0)
            argparse_control = _get_phono3py_load_args(
                cwd / ".." / "phono3py_params_Si-111-222.yaml",
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            argparse_control = _get_phono3py_load_args(
                "phono3py.yaml",
                is_lbte=True,
                is_wigner_kappa=True,
                temperatures=[
                    "300",
                ],
                mesh_numbers=["5", "5", "5"],
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            # for filename in pathlib.Path.cwd().iterdir():
            #     print(filename)

            # Clean files created by phono3py-load script.
            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "kappa-m555.hdf5",
                "coleigs-m555.hdf5",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

        finally:
            os.chdir(original_cwd)


@pytest.mark.parametrize(
    "load_phono3py_yaml,fc_calculator,fc_calculator_options",
    [
        (True, None, None),
        (True, "symfc", None),
        (True, "symfc", "|cutoff=4.0"),
        (False, "symfc", "|cutoff=4.0"),
    ],
)
def test_phono3py_load_with_typeII_dataset(
    fc_calculator: str | None,
    fc_calculator_options: str | None,
    load_phono3py_yaml: bool,
):
    """Test phono3py-load script with typeII dataset.

    When None, fallback to symfc.

    """
    pytest.importorskip("symfc")

    with tempfile.TemporaryDirectory() as temp_dir:
        original_cwd = pathlib.Path.cwd()
        os.chdir(temp_dir)

        try:
            argparse_control = _get_phono3py_load_args(
                cwd / ".." / "phono3py_params-Si111-rd.yaml.xz",
                load_phono3py_yaml=load_phono3py_yaml,
                fc_calculator=fc_calculator,
                fc_calculator_options=fc_calculator_options,
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            # Clean files created by phono3py-load script.
            for created_filename in ("phono3py.yaml", "fc2.hdf5", "fc3.hdf5"):
                file_path = pathlib.Path(created_filename)
                if file_path.exists():
                    if created_filename == "fc3.hdf5":
                        with h5py.File(file_path, "r") as f:
                            if fc_calculator_options is None:
                                assert "fc3_nonzero_indices" not in f
                            else:
                                assert "fc3_nonzero_indices" in f
                                assert "fc3_cutoff" in f
                                assert f["fc3_cutoff"][()] == pytest.approx(4.0)
                    file_path.unlink()

        finally:
            os.chdir(original_cwd)


@pytest.mark.parametrize("load_phono3py_yaml", [True, False])
def test_phono3py_with_QE_calculator(load_phono3py_yaml):
    """Test phono3py-load script with QE calculator."""
    with tempfile.TemporaryDirectory() as temp_dir:
        original_cwd = pathlib.Path.cwd()
        os.chdir(temp_dir)

        try:
            argparse_control = _get_phono3py_load_args(
                cwd / "phono3py_params-qe-Si222.yaml.xz",
                load_phono3py_yaml=load_phono3py_yaml,
                is_bterta=True,
                temperatures=[
                    "300",
                ],
                mesh_numbers=["11", "11", "11"],
            )
            with pytest.raises(SystemExit):
                main(**argparse_control)

            with h5py.File("kappa-m111111.hdf5", "r") as f:
                np.testing.assert_almost_equal(f["kappa"][0, 0], 118.93, decimal=1)

            # Clean files created by phono3py/phono3py-load script.
            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "kappa-m111111.hdf5",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

        finally:
            os.chdir(original_cwd)


def test_phono3py_load_with_pypolymlp_si():
    """Test phono3py-load script with pypolymlp.

    First run generates polymlp.yaml.
    Second run uses polymlp.yaml.

    """
    pytest.importorskip("pypolymlp", minversion="0.9.2")
    pytest.importorskip("symfc")

    with tempfile.TemporaryDirectory() as temp_dir:
        original_cwd = pathlib.Path.cwd()
        os.chdir(temp_dir)

        try:
            # Create fc2.hdf5
            argparse_control = _get_phono3py_load_args(
                cwd / ".." / "phono3py_params_Si-111-222-rd.yaml.xz",
                fc_calculator="symfc",
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0
            for created_filename in ("phono3py.yaml", "fc2.hdf5", "fc3.hdf5"):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
            pathlib.Path("fc3.hdf5").unlink()

            # Create MLP (polymlp.yaml)
            argparse_control = _get_phono3py_load_args(
                cwd / ".." / "phono3py_params_Si-111-222-rd.yaml.xz",
                use_pypolymlp=True,
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0
            for created_filename in ("phono3py.yaml", "polymlp.yaml"):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()

            # Create phono3py_mlp_eval_dataset.yaml
            argparse_control = _get_phono3py_load_args(
                "phono3py.yaml",
                fc_calculator="symfc",
                random_displacements="auto",
                use_pypolymlp=True,
            )

            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            ph3 = phono3py.load("phono3py_mlp_eval_dataset.yaml")
            assert len(ph3.displacements) == 8

            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "polymlp.yaml",
                "phono3py_mlp_eval_dataset.yaml",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

        finally:
            os.chdir(original_cwd)


def test_phono3py_load_with_pypolymlp_nacl():
    """Test phono3py-load script with pypolymlp using NaCl.

    First run generates polymlp.yaml.
    Second run uses polymlp.yaml.

    """
    pytest.importorskip("pypolymlp", minversion="0.9.2")
    pytest.importorskip("symfc")

    with tempfile.TemporaryDirectory() as temp_dir:
        original_cwd = pathlib.Path.cwd()
        os.chdir(temp_dir)

        try:
            # Stage1 (preparation)
            argparse_control = _get_phono3py_load_args(
                cwd / ".." / "phono3py_params_MgO-222rd-444rd.yaml.xz",
                mlp_params="cutoff=4.0,gtinv_maxl=4 4,max_p=1,gtinv_order=2",
                fc_calculator="symfc",
                random_displacements="auto",
                use_pypolymlp=True,
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            ph3 = phono3py.load("phono3py_mlp_eval_dataset.yaml")
            assert len(ph3.displacements) == 32

            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "polymlp.yaml",
                "phono3py_mlp_eval_dataset.yaml",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()

            for created_filename in (
                "fc3.hdf5",
                "phono3py_mlp_eval_dataset.yaml",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

            # Stage2 (cutoff test)
            argparse_control = _get_phono3py_load_args(
                "phono3py.yaml",
                fc_calculator="symfc",
                fc_calculator_options="|cutoff=4.0",
                random_displacements="auto",
                use_pypolymlp=True,
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            ph3 = phono3py.load("phono3py_mlp_eval_dataset.yaml")
            assert len(ph3.displacements) == 8

            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "polymlp.yaml",
                "phono3py_mlp_eval_dataset.yaml",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()

            for created_filename in (
                "fc3.hdf5",
                "phono3py_mlp_eval_dataset.yaml",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

            # Stage3 (memsize test)
            argparse_control = _get_phono3py_load_args(
                "phono3py.yaml",
                fc_calculator="symfc",
                fc_calculator_options="|memsize=0.05",
                random_displacements="auto",
                use_pypolymlp=True,
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            ph3 = phono3py.load("phono3py_mlp_eval_dataset.yaml")
            assert len(ph3.displacements) == 16

            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "polymlp.yaml",
                "phono3py_mlp_eval_dataset.yaml",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()

            for created_filename in (
                "fc3.hdf5",
                "phono3py_mlp_eval_dataset.yaml",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

            # Stage4 (number_estimation_factor)
            argparse_control = _get_phono3py_load_args(
                "phono3py.yaml",
                fc_calculator="symfc",
                fc_calculator_options="|cutoff=4.0",
                random_displacements="auto",
                rd_number_estimation_factor=2.0,
                use_pypolymlp=True,
            )
            with pytest.raises(SystemExit) as excinfo:
                main(**argparse_control)
            assert excinfo.value.code == 0

            ph3 = phono3py.load("phono3py_mlp_eval_dataset.yaml")
            assert len(ph3.displacements) == 4

            for created_filename in (
                "phono3py.yaml",
                "fc2.hdf5",
                "fc3.hdf5",
                "polymlp.yaml",
                "phono3py_mlp_eval_dataset.yaml",
            ):
                file_path = pathlib.Path(created_filename)
                assert file_path.exists()
                file_path.unlink()

        finally:
            os.chdir(original_cwd)


def _get_phono3py_load_args(
    phono3py_yaml_filepath: str | os.PathLike | None,
    fc_calculator: str | None = None,
    fc_calculator_options: str | None = None,
    load_phono3py_yaml: bool = True,
    is_bterta: bool = False,
    is_lbte: bool = False,
    is_wigner_kappa: bool = False,
    mesh_numbers: Sequence | None = None,
    mlp_params: str | None = None,
    rd_number_estimation_factor: float | None = None,
    random_displacements: int | str | None = None,
    temperatures: Sequence | None = None,
    use_pypolymlp: bool = False,
):
    # Mock of ArgumentParser.args.
    if load_phono3py_yaml:
        assert phono3py_yaml_filepath is not None
        mockargs = MockArgs(
            filename=[phono3py_yaml_filepath],
            fc_calculator=fc_calculator,
            fc_calculator_options=fc_calculator_options,
            is_bterta=is_bterta,
            is_lbte=is_lbte,
            is_wigner_kappa=is_wigner_kappa,
            log_level=1,
            mesh_numbers=mesh_numbers,
            mlp_params=mlp_params,
            rd_number_estimation_factor=rd_number_estimation_factor,
            random_displacements=random_displacements,
            temperatures=temperatures,
            use_pypolymlp=use_pypolymlp,
        )
    else:
        mockargs = MockArgs(
            filename=[],
            fc_calculator=fc_calculator,
            fc_calculator_options=fc_calculator_options,
            log_level=1,
            cell_filename=phono3py_yaml_filepath,
            is_bterta=is_bterta,
            is_lbte=is_lbte,
            is_wigner_kappa=is_wigner_kappa,
            mesh_numbers=mesh_numbers,
            mlp_params=mlp_params,
            rd_number_estimation_factor=rd_number_estimation_factor,
            random_displacements=random_displacements,
            temperatures=temperatures,
            use_pypolymlp=use_pypolymlp,
        )

    # See phono3py-load script.
    argparse_control = {
        "fc_symmetry": load_phono3py_yaml,
        "is_nac": load_phono3py_yaml,
        "load_phono3py_yaml": load_phono3py_yaml,
        "args": mockargs,
    }
    return argparse_control



"""Tests of Phono3py API."""

import os
import pathlib
import tempfile
from collections.abc import Sequence
from typing import Optional

import h5py
import numpy as np

from phono3py import Phono3py
from phono3py.file_IO import _get_filename_suffix

cwd = pathlib.Path(__file__).parent


def test_kappa_filename():
    """Test _get_filename_suffix."""
    mesh = [4, 4, 4]
    grid_point = None
    band_indices = None
    sigma = None
    sigma_cutoff = None
    filename = None
    suffix = _get_filename_suffix(
        mesh,
        grid_point=grid_point,
        band_indices=band_indices,
        sigma=sigma,
        sigma_cutoff=sigma_cutoff,
        filename=filename,
    )
    full_filename = "kappa" + suffix + ".hdf5"
    assert full_filename == "kappa-m444.hdf5"


def test_kappa_hdf5_with_boundary_mpf(si_pbesol: Phono3py):
    """Test boundary_mfp in kappa-*.hdf5.

    Remember to clean files created by
    Phono3py.run_thermal_conductivity(write_kappa=True).

    """
    key_ref = [
        "boundary_mfp",
        "frequency",
        "gamma",
        "grid_point",
        "group_velocity",
        "gv_by_gv",
        "heat_capacity",
        "kappa",
        "kappa_unit_conversion",
        "mesh",
        "mode_kappa",
        "qpoint",
        "temperature",
        "version",
        "weight",
    ]

    boundary_mfp = 10000.0

    with tempfile.TemporaryDirectory() as temp_dir:
        original_cwd = pathlib.Path.cwd()
        os.chdir(temp_dir)

        try:
            kappa_filename = _set_kappa(
                si_pbesol, [4, 4, 4], write_kappa=True, boundary_mfp=boundary_mfp
            )
            file_path = pathlib.Path(kappa_filename)
            with h5py.File(file_path, "r") as f:
                np.testing.assert_almost_equal(f["boundary_mfp"][()], boundary_mfp)
                assert set(list(f)) == set(key_ref)

            if file_path.exists():
                file_path.unlink()

        finally:
            os.chdir(original_cwd)


def _set_kappa(
    ph3: Phono3py,
    mesh: Sequence,
    is_isotope: bool = False,
    is_full_pp: bool = False,
    write_kappa: bool = False,
    boundary_mfp: Optional[float] = None,
) -> str:
    ph3.mesh_numbers = mesh
    ph3.init_phph_interaction()
    ph3.run_thermal_conductivity(
        temperatures=[
            300,
        ],
        is_isotope=is_isotope,
        is_full_pp=is_full_pp,
        write_kappa=write_kappa,
        boundary_mfp=boundary_mfp,
    )
    suffix = _get_filename_suffix(mesh)
    return "kappa" + suffix + ".hdf5"



"""Phono3py main command line script."""

# Copyright (C) 2020 Atsushi Togo
# All rights reserved.
#
# This file is part of phono3py.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution.
#
# * Neither the name of the phonopy project nor the names of its
#   contributors may be used to endorse or promote products derived
#   from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import annotations

import argparse
import os
import pathlib
import sys
from collections.abc import Sequence
from typing import cast

import numpy as np
from numpy.typing import ArrayLike, NDArray
from phonopy.api_phonopy import Phonopy
from phonopy.cui.phonopy_argparse import show_deprecated_option_warnings
from phonopy.cui.phonopy_script import (
    file_exists,
    print_end,
    print_error,
    print_error_message,
    print_time,
    print_version,
    store_nac_params,
)
from phonopy.cui.settings import PhonopySettings
from phonopy.exception import (
    CellNotFoundError,
    ForceCalculatorRequiredError,
    PypolymlpRelaxationError,
)
from phonopy.file_IO import is_file_phonopy_yaml
from phonopy.harmonic.dynamical_matrix import DynamicalMatrixGL
from phonopy.harmonic.force_constants import show_drift_force_constants
from phonopy.interface.calculator import get_calculator_physical_units
from phonopy.interface.pypolymlp import get_change_in_positions, relax_atomic_positions
from phonopy.interface.symfc import estimate_symfc_cutoff_from_memsize
from phonopy.phonon.band_structure import get_band_qpoints
from phonopy.physical_units import get_physical_units
from phonopy.structure.atoms import PhonopyAtoms
from phonopy.structure.cells import isclose as cells_isclose

from phono3py import Phono3py, Phono3pyIsotope, Phono3pyJointDos
from phono3py.cui.create_force_constants import (
    develop_pypolymlp,
    generate_displacements_and_evaluate_pypolymlp,
)
from phono3py.cui.create_force_sets import (
    create_FORCE_SETS_from_FORCES_FCx,
    create_FORCES_FC2_from_FORCE_SETS,
    create_FORCES_FC3_and_FORCES_FC2,
)
from phono3py.cui.create_supercells import (
    Phono3pyCellInfoResult,
    create_phono3py_supercells,
    get_cell_info,
)
from phono3py.cui.load import (
    compute_force_constants_from_datasets,
    load_fc2_and_fc3,
    select_and_load_dataset,
    select_and_load_phonon_dataset,
)
from phono3py.cui.phono3py_argparse import get_parser
from phono3py.cui.settings import Phono3pyConfParser, Phono3pySettings
from phono3py.cui.show_log import (
    show_general_settings,
    show_phono3py_cells,
    show_phono3py_settings,
)
from phono3py.cui.triplets_info import show_num_triplets, write_grid_points
from phono3py.file_IO import (
    read_phonon_from_hdf5,
    write_fc2_to_hdf5,
    write_fc3_to_hdf5,
    write_phonon_to_hdf5,
)
from phono3py.interface.fc_calculator import (
    determine_cutoff_pair_distance,
    get_fc_calculator_params,
)
from phono3py.interface.phono3py_yaml import Phono3pyYaml
from phono3py.phonon.grid import BZGrid, get_grid_point_from_address, get_ir_grid_points
from phono3py.phonon3.dataset import forces_in_dataset
from phono3py.phonon3.fc3 import show_drift_fc3
from phono3py.phonon3.gruneisen import run_gruneisen_parameters
from phono3py.version import __version__

# import logging
# logging.basicConfig()
# logging.getLogger("phono3py.phonon3.fc3").setLevel(level=logging.DEBUG)


# AA is created at http://www.network-science.de/ascii/.
def print_phono3py():
    """Show phono3py logo."""
    print(
        r"""        _                      _____
  _ __ | |__   ___  _ __   ___|___ / _ __  _   _
 | '_ \| '_ \ / _ \| '_ \ / _ \ |_ \| '_ \| | | |
 | |_) | | | | (_) | | | | (_) |__) | |_) | |_| |
 | .__/|_| |_|\___/|_| |_|\___/____/| .__/ \__, |
 |_|                                |_|    |___/ """
    )
    print_version(__version__, package_name="phono3py")
    print_time()


def print_end_phono3py():
    """Print END logo."""
    print_time()
    print_end()


def _finalize_phono3py(
    phono3py: Phono3py,
    confs_dict: dict,
    log_level: int,
    write_displacements: bool = False,
    filename: str | None = None,
):
    """Write phono3py.yaml and then exit.

    Parameters
    ----------
    phono3py : Phono3py
        Phono3py instance.
    confs_dict : dict
        This contains the settings and command options that the user set.
    log_level : int
        Log level. 0 means quiet.
    write_displacements : Bool
        When True, crystal structure is written in the length unit of
        calculator interface in phono3py_disp.yaml. Otherwise, the
        default unit (angstrom) is used.
    filename : str, optional
        phono3py.yaml is written in this filename.

    """
    if filename is None:
        yaml_filename = "phono3py.yaml"
    else:
        yaml_filename = filename

    _physical_units = get_calculator_physical_units(phono3py.calculator)

    ph3py_yaml = Phono3pyYaml(
        configuration=confs_dict,
        calculator=phono3py.calculator,
        physical_units=_physical_units,
        settings={
            "force_sets": False,
            "displacements": write_displacements,
        },
    )
    ph3py_yaml.set_phonon_info(phono3py)
    with open(yaml_filename, "w") as w:
        w.write(str(ph3py_yaml))

    if log_level > 0:
        if write_displacements:
            print(f'Displacement dataset was written in "{yaml_filename}".')
        else:
            print(f'Summary of calculation was written in "{yaml_filename}".')
        print_end_phono3py()
    sys.exit(0)


def _get_run_mode(settings: Phono3pySettings):
    """Extract run mode from settings."""
    if settings.is_gruneisen:
        run_mode = "gruneisen"
    elif settings.is_joint_dos:
        run_mode = "jdos"
    elif settings.is_isotope and not (settings.is_bterta or settings.is_lbte):
        run_mode = "isotope"
    elif settings.is_imag_self_energy:
        run_mode = "imag_self_energy"
    elif settings.is_real_self_energy:
        run_mode = "real_self_energy"
    elif settings.is_spectral_function:
        run_mode = "spectral_function"
    elif settings.is_bterta:
        run_mode = "conductivity-RTA"
    elif settings.is_lbte:
        run_mode = "conductivity-LBTE"
    elif (
        settings.create_displacements or settings.random_displacements is not None
    ) and not settings.use_pypolymlp:
        run_mode = "displacements"
    elif settings.write_phonon:
        run_mode = "phonon"
    else:
        run_mode = "force constants"
    return run_mode


def _start_phono3py(**argparse_control) -> tuple[argparse.Namespace, int]:
    """Parse arguments and set some basic parameters."""
    parser, deprecated = get_parser(argparse_control.get("load_phono3py_yaml", False))
    args = parser.parse_args()

    # Log level
    log_level = 1
    if args.verbose:
        log_level = 2
    if args.quiet:
        log_level = 0
    if args.log_level is not None:
        log_level = args.log_level

    # Title
    if log_level:
        print_phono3py()
        import phono3py._phono3py as phono3c  # type: ignore[import]

        max_threads = phono3c.omp_max_threads()
        if max_threads > 0:
            print(f"Compiled with OpenMP support (max {max_threads} threads).")
        if phono3c.include_lapacke():
            print("Compiled with LAPACKE.")

        if argparse_control.get("load_phono3py_yaml", False):
            print("Running in phono3py.load mode.")
        print("Python version %d.%d.%d" % sys.version_info[:3])
        import spglib

        try:  # spglib.get_version() is deprecated.
            print(f"Spglib version {spglib.spg_get_version()}")  # type: ignore
        except AttributeError:
            print("Spglib version %d.%d.%d" % spglib.get_version())  # type: ignore

        if deprecated:
            show_deprecated_option_warnings(deprecated)

    return args, log_level


def _read_phono3py_settings(
    args: argparse.Namespace, argparse_control: dict, log_level: int
):
    """Read phono3py settings.

    From:
    * Traditional configuration file.
    * phono3py.yaml type file
    * Command line options

    """
    load_phono3py_yaml = argparse_control.get("load_phono3py_yaml", False)

    if len(args.filename) > 0:
        file_exists(args.filename[0], log_level=log_level)
        if load_phono3py_yaml:
            phono3py_conf_parser = Phono3pyConfParser(
                filename=args.conf_filename,
                args=args,
                load_phono3py_yaml=load_phono3py_yaml,
            )
            cell_filename = args.filename[0]
        else:
            if is_file_phonopy_yaml(args.filename[0], keyword="phono3py"):
                phono3py_conf_parser = Phono3pyConfParser(
                    args=args, load_phono3py_yaml=load_phono3py_yaml
                )
                cell_filename = args.filename[0]
            else:  # args.filename[0] is assumed to be phono3py-conf file.
                phono3py_conf_parser = Phono3pyConfParser(
                    filename=args.filename[0],
                    args=args,
                    load_phono3py_yaml=load_phono3py_yaml,
                )
                cell_filename = phono3py_conf_parser.settings.cell_filename
    else:
        if load_phono3py_yaml:
            phono3py_conf_parser = Phono3pyConfParser(
                args=args,
                filename=args.conf_filename,
                load_phono3py_yaml=load_phono3py_yaml,
            )
        else:
            phono3py_conf_parser = Phono3pyConfParser(
                args=args, load_phono3py_yaml=load_phono3py_yaml
            )
        cell_filename = phono3py_conf_parser.settings.cell_filename

    confs_dict = phono3py_conf_parser.confs.copy()
    settings = phono3py_conf_parser.settings

    return settings, confs_dict, cell_filename


def _get_default_values(settings: Phono3pySettings):
    """Set default values."""
    # Brillouin zone integration: Tetrahedron (default) or smearing method
    sigma = settings.sigma
    if sigma is None:
        sigmas = []
    elif isinstance(sigma, float):
        sigmas = [sigma]
    else:
        sigmas = sigma
    if settings.is_tetrahedron_method:
        sigmas = [None] + sigmas
    if len(sigmas) == 0:
        sigmas = [None]

    if settings.temperatures is None:
        if settings.is_joint_dos:
            temperature_points = None
            temperatures = None
        else:
            t_max = settings.max_temperature
            t_min = settings.min_temperature
            t_step = settings.temperature_step
            temperature_points = [0.0, 300.0]  # For spectra
            temperatures = np.arange(t_min, t_max + float(t_step) / 10, t_step)
    else:
        temperature_points = settings.temperatures  # For spectra
        temperatures = settings.temperatures  # For others

    if settings.frequency_conversion_factor is None:
        frequency_factor_to_THz = get_physical_units().DefaultToTHz
    else:
        frequency_factor_to_THz = settings.frequency_conversion_factor

    if settings.num_frequency_points is None:
        if settings.frequency_pitch is None:
            num_frequency_points = 201
            frequency_step = None
        else:
            num_frequency_points = None
            frequency_step = settings.frequency_pitch
    else:
        num_frequency_points = settings.num_frequency_points
        frequency_step = None

    if settings.num_points_in_batch is None:
        num_points_in_batch = 10
    else:
        num_points_in_batch = settings.num_points_in_batch

    if settings.frequency_scale_factor is None:
        frequency_scale_factor = None
    else:
        frequency_scale_factor = settings.frequency_scale_factor

    if settings.cutoff_frequency is None:
        cutoff_frequency = 1e-2
    else:
        cutoff_frequency = settings.cutoff_frequency

    params = {}
    params["sigmas"] = sigmas
    params["temperature_points"] = temperature_points
    params["temperatures"] = temperatures
    params["frequency_factor_to_THz"] = frequency_factor_to_THz
    params["num_frequency_points"] = num_frequency_points
    params["num_points_in_batch"] = num_points_in_batch
    params["frequency_step"] = frequency_step
    params["frequency_scale_factor"] = frequency_scale_factor
    params["cutoff_frequency"] = cutoff_frequency

    return params


def _check_supercell_in_yaml(
    cell_info: Phono3pyCellInfoResult,
    ph3: Phono3py,
    distance_to_A: float | None,
    log_level: int,
):
    """Check consistency between generated cells and cells in yaml."""
    if cell_info.phono3py_yaml is not None:
        if distance_to_A is None:
            d2A = 1.0
        else:
            d2A = distance_to_A
        phono3py_yaml = cell_info.phono3py_yaml
        if phono3py_yaml.supercell is not None and ph3.supercell is not None:  # noqa E129
            yaml_cell = phono3py_yaml.supercell.copy()
            yaml_cell.cell = yaml_cell.cell * d2A
            if not cells_isclose(yaml_cell, ph3.supercell):
                if log_level:
                    print(
                        "Generated supercell is inconsistent with "
                        f'that in "{cell_info.optional_structure_info[0]}".'
                    )
                    print_error()
                sys.exit(1)
        if (
            phono3py_yaml.phonon_supercell is not None
            and ph3.phonon_supercell is not None
        ):  # noqa E129
            yaml_cell = phono3py_yaml.phonon_supercell.copy()
            yaml_cell.cell = yaml_cell.cell * d2A
            if not cells_isclose(yaml_cell, ph3.phonon_supercell):
                if log_level:
                    print(
                        "Generated phonon supercell is inconsistent with "
                        f'that in "{cell_info.optional_structure_info[0]}".'
                    )
                    print_error()
                sys.exit(1)


def _init_phono3py_with_cell_info(
    settings: Phono3pySettings,
    cell_info: Phono3pyCellInfoResult,
    interface_mode: str | None,
    symprec: float,
    log_level: int,
) -> tuple[Phono3py, dict]:
    """Initialize phono3py and update settings by default values."""
    # updated_settings keys
    # ('sigmas', 'temperature_points', 'temperatures',
    #  'frequency_factor_to_THz', 'num_frequency_points',
    #  'frequency_step', 'frequency_scale_factor',
    #  'cutoff_frequency')
    phono3py, updated_settings, distance_to_A = _init_phono3py(
        settings,
        cell_info.unitcell.copy(),
        supercell_matrix=cell_info.supercell_matrix,
        primitive_matrix=cell_info.primitive_matrix,
        phonon_supercell_matrix=cell_info.phonon_supercell_matrix,
        interface_mode=interface_mode,
        symprec=symprec,
        log_level=log_level,
    )
    _check_supercell_in_yaml(cell_info, phono3py, distance_to_A, log_level)

    return phono3py, updated_settings


def _init_phono3py(
    settings: Phono3pySettings,
    unitcell: PhonopyAtoms,
    supercell_matrix: ArrayLike | None = None,
    primitive_matrix: ArrayLike | None = None,
    phonon_supercell_matrix: ArrayLike | None = None,
    interface_mode: str | None = None,
    symprec: float = 1e-5,
    log_level: int = 0,
) -> tuple[Phono3py, dict, float | None]:
    """Initialize phono3py and update settings by default values."""
    if interface_mode is not None:
        physical_units = get_calculator_physical_units(interface_mode)
        distance_to_A = physical_units["distance_to_A"]
        assert distance_to_A is not None
        lattice = unitcell.cell
        lattice *= distance_to_A
        unitcell.cell = lattice
    else:
        distance_to_A = None

    updated_settings = _get_default_values(settings)

    phono3py = Phono3py(
        unitcell,
        supercell_matrix,
        primitive_matrix=primitive_matrix,
        phonon_supercell_matrix=phonon_supercell_matrix,
        cutoff_frequency=updated_settings["cutoff_frequency"],
        frequency_factor_to_THz=updated_settings["frequency_factor_to_THz"],
        is_symmetry=settings.is_symmetry,
        is_mesh_symmetry=settings.is_mesh_symmetry,
        use_grg=settings.use_grg,
        make_r0_average=settings.is_fc3_r0_average,
        symprec=symprec,
        log_level=log_level,
    )
    phono3py.masses = settings.masses
    phono3py.band_indices = settings.band_indices
    phono3py.sigmas = updated_settings["sigmas"]
    phono3py.sigma_cutoff = settings.sigma_cutoff_width

    return phono3py, updated_settings, distance_to_A


def _settings_to_grid_points(
    settings: Phono3pySettings, bz_grid: BZGrid
) -> ArrayLike | None:
    """Read or set grid point indices."""
    if settings.grid_addresses is not None:
        grid_points = _grid_addresses_to_grid_points(settings.grid_addresses, bz_grid)
    elif settings.grid_points is not None:
        grid_points = settings.grid_points
    else:
        grid_points = None
    return grid_points


def _grid_addresses_to_grid_points(grid_addresses: NDArray, bz_grid: BZGrid) -> NDArray:
    """Return grid point indices from grid addresses."""
    grid_points = [
        get_grid_point_from_address(ga, bz_grid.D_diag) for ga in grid_addresses
    ]
    return bz_grid.grg2bzg[grid_points]


def _create_supercells_with_displacements(
    settings: Phono3pySettings,
    cell_info: Phono3pyCellInfoResult,
    confs_dict: dict,
    unitcell_filename: str,
    interface_mode: str | None,
    symprec: float = 1e-5,
    output_yaml_filename: str | None = None,
    log_level: int = 0,
):
    """Create supercells and write displacements."""
    if (
        settings.create_displacements
        or settings.random_displacements is not None
        or settings.random_displacements_fc2 is not None
    ):
        ph3py = create_phono3py_supercells(
            cell_info,
            settings,
            symprec,
            interface_mode=interface_mode,
            log_level=log_level,
        )

        if pathlib.Path("BORN").exists():
            store_nac_params(
                cast(Phonopy, ph3py),
                cast(PhonopySettings, settings),
                cell_info.phono3py_yaml,
                unitcell_filename,
                log_level,
                nac_factor=get_physical_units().Hartree * get_physical_units().Bohr,
            )

        if log_level:
            if ph3py.supercell.magnetic_moments is None:
                print("Spacegroup: %s" % ph3py.symmetry.get_international_table())
            else:
                print(
                    "Number of symmetry operations in supercell: %d"
                    % len(ph3py.symmetry.symmetry_operations["rotations"])
                )

        ph3py.save("phono3py_disp.yaml")
        if log_level > 0:
            print('Displacement dataset was written in "phono3py_disp.yaml".')

        _finalize_phono3py(
            ph3py,
            confs_dict,
            log_level,
            filename=output_yaml_filename,
        )


def _run_pypolymlp(
    ph3py: Phono3py,
    settings: Phono3pySettings,
    confs_dict: dict,
    output_yaml_filename: str | None = None,
    mlp_eval_filename: str = "phono3py_mlp_eval_dataset.yaml",
    log_level: int = 0,
) -> Phono3py:
    assert ph3py.mlp_dataset is None
    if ph3py.dataset is not None:  # If None, load mlp from polymlp.yaml.
        ph3py.mlp_dataset = ph3py.dataset
        ph3py.dataset = None

    develop_pypolymlp(
        ph3py,
        mlp_params=settings.mlp_params,
        log_level=log_level,
    )

    _ph3py = ph3py
    if settings.relax_atomic_positions:
        if log_level:
            print("Relaxing atomic positions using polynomial MLPs...")

        try:
            assert ph3py.mlp is not None
            relaxed_unitcell = relax_atomic_positions(
                ph3py.unitcell,
                ph3py.mlp.mlp,
                verbose=log_level > 1,
            )
        except (PypolymlpRelaxationError, ValueError) as e:
            # ValueError can come from pypolymlp directly.
            print_error_message(str(e))
            if log_level:
                print_error()
            sys.exit(1)

        if log_level:
            if relaxed_unitcell is None:
                print("No relaxation was performed due to symmetry constraints.")
            else:
                get_change_in_positions(
                    relaxed_unitcell, ph3py.unitcell, verbose=log_level > 0
                )
            print("-" * 76)

        if relaxed_unitcell is not None:
            if log_level:
                print(
                    "Relaxed crystal structures will be stored in "
                    f'"{output_yaml_filename}".'
                )
            _ph3py, _, _ = _init_phono3py(
                settings,
                relaxed_unitcell.copy(),
                supercell_matrix=ph3py.supercell_matrix,
                primitive_matrix=ph3py.primitive_matrix,
                phonon_supercell_matrix=ph3py.phonon_supercell_matrix,
                symprec=ph3py.symmetry.tolerance,
                log_level=log_level,
            )
            if ph3py.mesh_numbers is not None:
                assert settings.mesh_numbers is not None
                _ph3py.mesh_numbers = settings.mesh_numbers
            _ph3py.nac_params = ph3py.nac_params
            _ph3py.mlp = ph3py.mlp

    if settings.create_displacements or settings.random_displacements is not None:
        generate_displacements_and_evaluate_pypolymlp(
            _ph3py,
            displacement_distance=settings.displacement_distance,
            number_of_snapshots=settings.random_displacements,
            number_estimation_factor=settings.rd_number_estimation_factor,
            random_seed=settings.random_seed,
            fc_calculator=settings.fc_calculator,
            fc_calculator_options=settings.fc_calculator_options,
            cutoff_pair_distance=settings.cutoff_pair_distance,
            symfc_memory_size=settings.symfc_memory_size,
            log_level=log_level,
        )
        if log_level:
            print(f'Dataset generated using MLPs was written in "{mlp_eval_filename}".')
        _ph3py.save(mlp_eval_filename)
    else:
        if log_level:
            print(
                "Generate displacements (--rd or -d) for proceeding to phonon "
                "calculations."
            )
        _finalize_phono3py(
            _ph3py, confs_dict, log_level=log_level, filename=output_yaml_filename
        )

    return _ph3py


def _produce_force_constants(
    ph3py: Phono3py,
    settings: Phono3pySettings,
    log_level: int,
    load_phono3py_yaml: bool,
):
    """Calculate, read, and write force constants."""
    if log_level:
        print("-" * 29 + " Force constants " + "-" * 30)

    read_fc3 = ph3py.fc3 is not None
    read_fc2 = ph3py.fc2 is not None

    cutoff_pair_distance = None
    if settings.use_pypolymlp and ph3py.dataset is not None:
        cutoff_pair_distance = ph3py.dataset.get("cutoff_distance")
    if cutoff_pair_distance is None:
        cutoff_pair_distance = determine_cutoff_pair_distance(
            fc_calculator=settings.fc_calculator,
            fc_calculator_options=settings.fc_calculator_options,
            cutoff_pair_distance=settings.cutoff_pair_distance,
            symfc_memory_size=settings.symfc_memory_size,
            random_displacements=settings.random_displacements,
            supercell=ph3py.supercell,
            primitive=ph3py.primitive,
            log_level=log_level,
        )
    if cutoff_pair_distance is None and ph3py.dataset is not None:
        cutoff_pair_distance = ph3py.dataset.get("cutoff_distance")

    (fc_calculator, fc_calculator_options) = get_fc_calculator_params(
        settings.fc_calculator,
        settings.fc_calculator_options,
        cutoff_pair_distance,
        log_level=(not read_fc3) * 1,
    )
    try:
        compute_force_constants_from_datasets(
            ph3py,
            fc_calculator=fc_calculator,
            fc_calculator_options=fc_calculator_options,
            cutoff_pair_distance=cutoff_pair_distance,
            symmetrize_fc=settings.fc_symmetry,
            is_compact_fc=settings.is_compact_fc,
            load_phono3py_yaml=load_phono3py_yaml,
        )
    except ForceCalculatorRequiredError as e:
        if load_phono3py_yaml:
            if log_level:
                print(str(e))
                print("Try symfc to handle general (or random) displacements.")
        else:
            print_error_message(str(e))
            if log_level:
                print_error()
            sys.exit(1)

        compute_force_constants_from_datasets(
            ph3py,
            fc_calculator="symfc",
            fc_calculator_options=fc_calculator_options,
            cutoff_pair_distance=cutoff_pair_distance,
            symmetrize_fc=settings.fc_symmetry,
            is_compact_fc=settings.is_compact_fc,
        )

    if log_level:
        if ph3py.fc3 is None:
            print("fc3 could not be obtained.")
            if not forces_in_dataset(ph3py.dataset):
                print("Forces were not found.")
        if ph3py.fc2 is None:
            print("fc2 could not be obtained.")
            if ph3py.phonon_supercell_matrix is None:
                if not forces_in_dataset(ph3py.dataset):
                    print("Forces were not found.")
            else:
                if not forces_in_dataset(ph3py.phonon_dataset):
                    print("Forces for dim-fc2 were not found.")

    if ph3py.fc3 is None or ph3py.fc2 is None:
        print_error()
        sys.exit(1)

    if log_level:
        show_drift_fc3(ph3py.fc3, primitive=ph3py.primitive)
        show_drift_force_constants(
            ph3py.fc2, primitive=ph3py.phonon_primitive, name="fc2"
        )

    cutoff_distance = settings.cutoff_fc3_distance
    if cutoff_distance is not None and cutoff_distance > 0:
        if log_level:
            print("Cutting-off fc3 by zero (cut-off distance: %f)" % cutoff_distance)
        ph3py.cutoff_fc3_by_zero(cutoff_distance)

    if not read_fc3:
        write_fc3_to_hdf5(
            ph3py.fc3,
            fc3_nonzero_indices=ph3py.fc3_nonzero_indices,
            p2s_map=ph3py.primitive.p2s_map,
            fc3_cutoff=ph3py.fc3_cutoff,
            compression=settings.hdf5_compression,
        )
        if log_level:
            print('fc3 was written into "fc3.hdf5".')
    if not read_fc2:
        write_fc2_to_hdf5(
            ph3py.fc2,
            p2s_map=ph3py.phonon_primitive.p2s_map,
            physical_unit="eV/angstrom^2",
            compression=settings.hdf5_compression,
        )
        if log_level:
            print('fc2 was written into "fc2.hdf5".')


def _run_gruneisen_then_exit(
    phono3py: Phono3py,
    settings: Phono3pySettings,
    log_level: int,
):
    """Run mode Grueneisen parameter calculation from fc3."""
    if (
        settings.mesh_numbers is None
        and settings.band_paths is None
        and settings.qpoints is None
    ):
        print("An option of --mesh, --band, or --qpoints has to be specified.")
        if log_level:
            print_error()
        sys.exit(1)

    assert phono3py.fc2 is not None
    assert phono3py.fc3 is not None
    if len(phono3py.fc2) != len(phono3py.fc3):
        print("Supercells used for fc2 and fc3 have to be same.")
        if log_level:
            print_error()
        sys.exit(1)

    if settings.band_paths is not None:
        if settings.band_points is None:
            npoints = 51
        else:
            npoints = settings.band_points
        band_paths = settings.band_paths
        bands = get_band_qpoints(band_paths, npoints=npoints)
    else:
        bands = None

    rotations = phono3py.primitive_symmetry.pointgroup_operations
    run_gruneisen_parameters(
        phono3py.fc2,
        phono3py.fc3,
        phono3py.supercell,
        phono3py.primitive,
        bands,
        settings.mesh_numbers,
        rotations,
        settings.qpoints,
        nac_params=phono3py.nac_params,
        nac_q_direction=settings.nac_q_direction,
        ion_clamped=settings.ion_clamped,
        factor=get_physical_units().DefaultToTHz,
        symprec=phono3py.symmetry.tolerance,
        log_level=log_level,
    )

    if log_level:
        print_end_phono3py()
    sys.exit(0)


def _run_jdos_then_exit(
    phono3py: Phono3py,
    settings: Phono3pySettings,
    updated_settings: dict,
    log_level: int,
):
    """Run joint-DOS calculation."""
    joint_dos = Phono3pyJointDos(
        phono3py.phonon_supercell,
        phono3py.phonon_primitive,
        phono3py.fc2,
        mesh=settings.mesh_numbers,
        nac_params=phono3py.nac_params,
        nac_q_direction=settings.nac_q_direction,
        sigmas=updated_settings["sigmas"],
        cutoff_frequency=updated_settings["cutoff_frequency"],
        frequency_step=updated_settings["frequency_step"],
        num_frequency_points=updated_settings["num_frequency_points"],
        num_points_in_batch=updated_settings["num_points_in_batch"],
        temperatures=updated_settings["temperature_points"],
        frequency_factor_to_THz=updated_settings["frequency_factor_to_THz"],
        frequency_scale_factor=updated_settings["frequency_scale_factor"],
        use_grg=settings.use_grg,
        is_mesh_symmetry=settings.is_mesh_symmetry,
        symprec=phono3py.symmetry.tolerance,
        log_level=log_level,
    )

    if log_level > 0:
        dm = joint_dos.dynamical_matrix
        if isinstance(dm, DynamicalMatrixGL):
            dm.show_nac_message()

    assert joint_dos.grid is not None
    grid_points = _settings_to_grid_points(settings, joint_dos.grid)
    joint_dos.run(grid_points, write_jdos=True)

    if log_level:
        print_end_phono3py()
    sys.exit(0)


def _run_isotope_then_exit(
    phono3py: Phono3py,
    settings: Phono3pySettings,
    updated_settings: dict,
    log_level: int,
):
    """Run isotope scattering calculation."""
    mass_variances = settings.mass_variances
    if settings.band_indices is not None:
        band_indices = np.hstack(settings.band_indices).astype("intc")
    else:
        band_indices = None
    iso = Phono3pyIsotope(
        settings.mesh_numbers,
        phono3py.phonon_primitive,
        mass_variances=mass_variances,
        band_indices=band_indices,
        sigmas=updated_settings["sigmas"],
        frequency_factor_to_THz=updated_settings["frequency_factor_to_THz"],
        use_grg=settings.use_grg,
        symprec=phono3py.symmetry.tolerance,
        cutoff_frequency=settings.cutoff_frequency,
        lapack_zheev_uplo=settings.lapack_zheev_uplo,
    )
    iso.init_dynamical_matrix(
        phono3py.fc2,
        phono3py.phonon_supercell,
        phono3py.phonon_primitive,
        nac_params=phono3py.nac_params,
        frequency_scale_factor=updated_settings["frequency_scale_factor"],
    )
    if log_level > 0:
        dm = iso.dynamical_matrix
        if isinstance(dm, DynamicalMatrixGL):
            dm.show_nac_message()

    grid_points = _settings_to_grid_points(settings, iso.grid)
    iso.run(grid_points)

    if log_level:
        print_end_phono3py()
    sys.exit(0)


def _init_phph_interaction(
    phono3py: Phono3py,
    settings: Phono3pySettings,
    updated_settings: dict,
    log_level: int,
):
    """Initialize ph-ph interaction and phonons on grid."""
    if log_level:
        print("Generating grid system ... ", end="", flush=True)
    assert phono3py.grid is not None
    assert phono3py.mesh_numbers is not None
    bz_grid = phono3py.grid
    if log_level:
        if bz_grid.grid_matrix is None:
            print("[ %d %d %d ]" % tuple(phono3py.mesh_numbers))
        else:
            print("")
            print("Generalized regular grid: [ %d %d %d ]" % tuple(bz_grid.D_diag))
            print("Grid generation matrix:")
            print("  [ %d %d %d ]" % tuple(bz_grid.grid_matrix[0]))
            print("  [ %d %d %d ]" % tuple(bz_grid.grid_matrix[1]))
            print("  [ %d %d %d ]" % tuple(bz_grid.grid_matrix[2]))

        if settings.is_symmetrize_fc3_q:
            print("Permutation symmetry of ph-ph interaction strengths: True")
        if settings.is_fc3_r0_average:
            print("fc3 r2q transformation over three atoms: True")

    ave_pp = settings.constant_averaged_pp_interaction
    phono3py.init_phph_interaction(
        nac_q_direction=settings.nac_q_direction,
        constant_averaged_interaction=ave_pp,
        frequency_scale_factor=updated_settings["frequency_scale_factor"],
        symmetrize_fc3q=settings.is_symmetrize_fc3_q,
        lapack_zheev_uplo=settings.lapack_zheev_uplo,
    )

    if not settings.read_phonon:
        if log_level:
            print("-" * 27 + " Phonon calculations " + "-" * 28)
            dm = phono3py.dynamical_matrix
            if isinstance(dm, DynamicalMatrixGL):
                dm.show_nac_message()
            print("Running harmonic phonon calculations...")
            sys.stdout.flush()
        phono3py.run_phonon_solver()

    if settings.write_phonon:
        freqs, eigvecs, grid_address = phono3py.get_phonon_data()
        ir_grid_points, ir_grid_weights, _ = get_ir_grid_points(bz_grid)
        ir_grid_points = np.array(bz_grid.grg2bzg[ir_grid_points], dtype="int64")
        filename = write_phonon_to_hdf5(
            freqs,
            eigvecs,
            grid_address,
            phono3py.mesh_numbers,
            bz_grid=bz_grid,
            ir_grid_points=ir_grid_points,
            ir_grid_weights=ir_grid_weights,
            compression=settings.hdf5_compression,
        )
        if filename:
            if log_level:
                print('Phonons are written into "%s".' % filename)
        else:
            print("Writing phonons failed.")
            if log_level:
                print_error()
            sys.exit(1)

    if settings.read_phonon:
        phonons = read_phonon_from_hdf5(phono3py.mesh_numbers, verbose=(log_level > 0))
        if phonons is None:
            print("Reading phonons failed.")
            if log_level:
                print_error()
            sys.exit(1)

        try:
            phono3py.set_phonon_data(*phonons)
        except RuntimeError:
            if log_level:
                print_error()
            sys.exit(1)


def _load_dataset_and_phonon_dataset(
    ph3py: Phono3py,
    ph3py_yaml: Phono3pyYaml | None = None,
    forces_fc3_filename: str | os.PathLike | Sequence | None = None,
    forces_fc2_filename: str | os.PathLike | Sequence | None = None,
    phono3py_yaml_filename: str | os.PathLike | None = None,
    cutoff_pair_distance: float | None = None,
    calculator: str | None = None,
    log_level: int = 0,
):
    """Set displacements, forces, and create force constants."""
    if ph3py.fc3 is None or (
        ph3py.fc2 is None and ph3py.phonon_supercell_matrix is None
    ):
        dataset = select_and_load_dataset(
            ph3py,
            ph3py_yaml=ph3py_yaml,
            forces_fc3_filename=forces_fc3_filename,
            phono3py_yaml_filename=phono3py_yaml_filename,
            cutoff_pair_distance=cutoff_pair_distance,
            calculator=calculator,
            log_level=log_level,
        )
        if dataset is not None:
            ph3py.dataset = dataset

    if ph3py.fc2 is None and ph3py.phonon_supercell_matrix is not None:
        phonon_dataset = select_and_load_phonon_dataset(
            ph3py,
            ph3py_yaml=ph3py_yaml,
            forces_fc2_filename=forces_fc2_filename,
            calculator=calculator,
            log_level=log_level,
        )
        if phonon_dataset is not None:
            ph3py.phonon_dataset = phonon_dataset


def main(**argparse_control):
    """Phono3py main part of command line interface."""
    # import warnings

    # warnings.simplefilter("error")
    load_phono3py_yaml = argparse_control.get("load_phono3py_yaml", False)

    if "args" in argparse_control:  # This is for pytest.
        args = argparse_control["args"]
        log_level = args.log_level
    else:
        args, log_level = _start_phono3py(**argparse_control)

    output_yaml_filename: str | None
    if load_phono3py_yaml:
        output_yaml_filename = args.output_yaml_filename
    else:
        output_yaml_filename = None

    settings, confs_dict, cell_filename = _read_phono3py_settings(
        args, argparse_control, log_level
    )

    if args.force_sets_to_forces_fc2_mode:
        create_FORCES_FC2_from_FORCE_SETS(log_level)
        if log_level:
            print_end_phono3py()
        sys.exit(0)
    if args.force_sets_mode:
        create_FORCE_SETS_from_FORCES_FCx(
            settings.phonon_supercell_matrix, cell_filename, log_level
        )
        if log_level:
            print_end_phono3py()
        sys.exit(0)
    if args.write_grid_points:
        run_mode = "write_grid_info"
    elif args.show_num_triplets:
        run_mode = "show_triplets_info"
    else:
        run_mode = None

    # -----------------------------------------------------------------------
    # ----------------- 'args' should not be used below. --------------------
    # -----------------------------------------------------------------------

    ####################################
    # Create FORCES_FC3 and FORCES_FC2 #
    ####################################
    if (
        settings.create_forces_fc3
        or settings.create_forces_fc3_file
        or settings.create_forces_fc2
    ):
        create_FORCES_FC3_and_FORCES_FC2(settings, cell_filename, log_level=log_level)
        if log_level:
            print_end_phono3py()
        sys.exit(0)

    ###########################################################
    # Symmetry tolerance. Distance unit depends on interface. #
    ###########################################################
    if settings.symmetry_tolerance is None:
        symprec = 1e-5
    else:
        symprec = settings.symmetry_tolerance

    try:
        cell_info = get_cell_info(
            settings,
            cell_filename,
            log_level=log_level,
            load_phonopy_yaml=load_phono3py_yaml,
        )
    except CellNotFoundError as e:
        print_error_message(str(e))
        if log_level:
            print_error()
        sys.exit(1)

    unitcell_filename = cell_info.optional_structure_info[0]
    interface_mode = cell_info.interface_mode

    if run_mode is None:
        run_mode = _get_run_mode(settings)
    assert run_mode is not None

    ######################################################
    # Create supercells with displacements and then exit #
    ######################################################
    if not settings.use_pypolymlp:
        _create_supercells_with_displacements(
            settings,
            cell_info,
            confs_dict,
            unitcell_filename,
            interface_mode,
            symprec=symprec,
            output_yaml_filename=output_yaml_filename,
            log_level=log_level,
        )

    #######################
    # Initialize phono3py #
    #######################
    # updated_settings keys
    # ('sigmas', 'temperature_points', 'temperatures',
    #  'frequency_factor_to_THz', 'num_frequency_points',
    #  'frequency_step', 'frequency_scale_factor',
    #  'cutoff_frequency')
    ph3py, updated_settings = _init_phono3py_with_cell_info(
        settings, cell_info, interface_mode, symprec, log_level
    )

    #################################################
    # Show phono3py settings and crystal structures #
    #################################################
    if log_level:
        show_general_settings(
            settings,
            run_mode,
            ph3py,
            unitcell_filename,
            interface_mode,
        )

        if ph3py.supercell.magnetic_moments is None:
            print("Spacegroup: %s" % ph3py.symmetry.get_international_table())
        else:
            print(
                "Number of symmetry operations in supercell: %d"
                % len(ph3py.symmetry.symmetry_operations["rotations"])
            )

    if log_level > 1:
        show_phono3py_cells(ph3py)
    elif log_level:
        print(
            "Use -v option to watch primitive cell, unit cell, "
            "and supercell structures."
        )

    ###############################
    # Memory estimation for symfc #
    ###############################
    if settings.show_symfc_memory_usage and load_phono3py_yaml:
        print("Quick estimation of memory size required for solving fc3 by symfc")
        print("cutoff   memsize")
        print("------   -------")
        estimate_symfc_cutoff_from_memsize(
            ph3py.supercell, ph3py.primitive, ph3py.symmetry, 3, verbose=True
        )

        if log_level:
            print_end_phono3py()
        sys.exit(0)

    ##################
    # Check settings #
    ##################
    run_modes_with_mesh = (
        "conductivity-RTA",
        "conductivity-LBTE",
        "imag_self_energy",
        "real_self_energy",
        "jdos",
        "isotope",
        "write_grid_info",
        "show_triplets_info",
    )
    run_modes_with_gp = ("imag_self_energy", "real_self_energy", "jdos", "isotope")

    if settings.mesh_numbers is None and run_mode in run_modes_with_mesh:
        print("")
        print("Mesh numbers have to be specified.")
        print("")
        if log_level:
            print_error()
        sys.exit(1)

    if (
        run_mode in run_modes_with_gp
        and settings.grid_points is None
        and settings.grid_addresses is None
    ):  # noqa E129
        print("")
        print("Grid point(s) has to be specified.")
        print("")
        if log_level:
            print_error()
        sys.exit(1)

    ####################
    # Set mesh numbers #
    ####################
    if run_mode in run_modes_with_mesh:
        assert settings.mesh_numbers is not None
        # jdos and isotope modes need to set mesh numbers differently.
        if run_mode not in ("jdos", "isotope"):
            ph3py.mesh_numbers = settings.mesh_numbers

    #########################################################
    # Write ir-grid points and grid addresses and then exit #
    #########################################################
    if run_mode == "write_grid_info":
        assert ph3py.grid is not None
        write_grid_points(
            ph3py.primitive,
            ph3py.grid,
            band_indices=settings.band_indices,
            sigmas=updated_settings["sigmas"],
            temperatures=updated_settings["temperatures"],
            is_kappa_star=settings.is_kappa_star,
            is_lbte=(settings.write_collision or settings.is_lbte),
            compression=settings.hdf5_compression,
        )

        if log_level:
            print_end_phono3py()
        sys.exit(0)

    ################################################################
    # Show reduced number of triplets at grid points and then exit #
    ################################################################
    if run_mode == "show_triplets_info":
        assert ph3py.grid is not None
        grid_points = _settings_to_grid_points(settings, ph3py.grid)
        show_num_triplets(
            ph3py.primitive,
            ph3py.grid,
            band_indices=settings.band_indices,
            grid_points=grid_points,
            is_kappa_star=settings.is_kappa_star,
        )

        if log_level:
            print_end_phono3py()
        sys.exit(0)

    ##################################
    # Non-analytical term correction #
    ##################################
    if settings.is_nac:
        store_nac_params(
            cast(Phonopy, ph3py),
            settings,
            cell_info.phono3py_yaml,
            unitcell_filename,
            log_level,
            nac_factor=get_physical_units().Hartree * get_physical_units().Bohr,
        )

    ########################
    # Read force constants #
    ########################
    load_fc2_and_fc3(
        ph3py,
        read_fc3=settings.read_fc3,
        read_fc2=settings.read_fc2,
        log_level=log_level,
    )

    ############
    # Datasets #
    ############
    _load_dataset_and_phonon_dataset(
        ph3py,
        ph3py_yaml=cell_info.phono3py_yaml,
        phono3py_yaml_filename=unitcell_filename,
        cutoff_pair_distance=settings.cutoff_pair_distance,
        calculator=interface_mode,
        log_level=log_level,
    )

    ###################
    # polynomial MLPs #
    ###################
    if settings.use_pypolymlp:
        if ph3py.fc3 is None or (
            ph3py.fc2 is None and ph3py.phonon_supercell_matrix is None
        ):
            # Note that ph3py is replaced when relax_atomic_positions is True.
            ph3py = _run_pypolymlp(
                ph3py,
                settings,
                confs_dict,
                output_yaml_filename=output_yaml_filename,
                log_level=log_level,
            )
        else:
            if log_level:
                print(
                    "Pypolymlp was not developed or used because fc2 and fc3 "
                    "are available."
                )

    ###########################
    # Produce force constants #
    ###########################
    _produce_force_constants(ph3py, settings, log_level, load_phono3py_yaml)

    ############################################
    # Phonon Gruneisen parameter and then exit #
    ############################################
    if settings.is_gruneisen:
        _run_gruneisen_then_exit(ph3py, settings, log_level)

    #################
    # Show settings #
    #################
    if log_level:
        show_phono3py_settings(ph3py, settings, updated_settings, log_level)

    ###########################
    # Joint DOS and then exit #
    ###########################
    if run_mode == "jdos":
        _run_jdos_then_exit(ph3py, settings, updated_settings, log_level)

    ################################################
    # Mass variances for phonon-isotope scattering #
    ################################################
    if settings.is_isotope and settings.mass_variances is None:
        from phonopy.structure.atoms import isotope_data

        symbols = ph3py.phonon_primitive.symbols
        in_database = True
        for s in set(symbols):
            if s not in isotope_data:
                print("%s is not in the list of isotope databese" % s)
                print("(not implemented).")
                print("Use --mass_variances option.")
                in_database = False
        if not in_database:
            if log_level:
                print_end_phono3py()
            sys.exit(0)

    #########################################
    # Phonon-isotope lifetime and then exit #
    #########################################
    if run_mode == "isotope":
        _run_isotope_then_exit(ph3py, settings, updated_settings, log_level)

    ########################################
    # Initialize phonon-phonon interaction #
    ########################################
    if run_mode in run_modes_with_mesh:
        _init_phph_interaction(
            ph3py,
            settings,
            updated_settings,
            log_level,
        )

    #######################################################
    # Run imaginary part of self energy of bubble diagram #
    #######################################################
    if run_mode == "imag_self_energy":
        assert ph3py.grid is not None
        ph3py.run_imag_self_energy(
            _settings_to_grid_points(settings, ph3py.grid),
            updated_settings["temperature_points"],
            frequency_step=updated_settings["frequency_step"],
            num_frequency_points=updated_settings["num_frequency_points"],
            num_points_in_batch=updated_settings["num_points_in_batch"],
            scattering_event_class=settings.scattering_event_class,
            write_txt=True,
            write_gamma_detail=settings.write_gamma_detail,
        )

    #####################################################
    # Run frequency shift calculation of bubble diagram #
    #####################################################
    elif run_mode == "real_self_energy":
        assert ph3py.grid is not None
        ph3py.run_real_self_energy(
            _settings_to_grid_points(settings, ph3py.grid),
            updated_settings["temperature_points"],
            frequency_step=updated_settings["frequency_step"],
            num_frequency_points=updated_settings["num_frequency_points"],
            write_txt=True,
            write_hdf5=True,
        )

    #######################################################
    # Run spectral function calculation of bubble diagram #
    #######################################################
    elif run_mode == "spectral_function":
        assert ph3py.grid is not None
        ph3py.run_spectral_function(
            _settings_to_grid_points(settings, ph3py.grid),
            updated_settings["temperature_points"],
            frequency_step=updated_settings["frequency_step"],
            num_frequency_points=updated_settings["num_frequency_points"],
            num_points_in_batch=updated_settings["num_points_in_batch"],
            write_txt=True,
            write_hdf5=True,
        )

    ####################################
    # Run lattice thermal conductivity #
    ####################################
    elif run_mode == "conductivity-RTA" or run_mode == "conductivity-LBTE":
        assert ph3py.grid is not None
        grid_points = _settings_to_grid_points(settings, ph3py.grid)
        ph3py.run_thermal_conductivity(
            is_LBTE=settings.is_lbte,
            temperatures=updated_settings["temperatures"],
            is_isotope=settings.is_isotope,
            mass_variances=settings.mass_variances,
            grid_points=grid_points,
            boundary_mfp=settings.boundary_mfp,
            solve_collective_phonon=settings.solve_collective_phonon,
            use_ave_pp=settings.use_ave_pp,
            is_reducible_collision_matrix=settings.is_reducible_collision_matrix,  # noqa E501
            is_kappa_star=settings.is_kappa_star,
            gv_delta_q=settings.group_velocity_delta_q,
            is_full_pp=(settings.is_full_pp or settings.is_symmetrize_fc3_q),
            pinv_cutoff=settings.pinv_cutoff,
            pinv_solver=settings.pinv_solver,
            pinv_method=settings.pinv_method,
            write_gamma=settings.write_gamma,
            read_gamma=settings.read_gamma,
            write_kappa=True,
            is_N_U=settings.is_N_U,
            conductivity_type=settings.conductivity_type,
            write_gamma_detail=settings.write_gamma_detail,
            write_collision=settings.write_collision,
            read_collision=settings.read_collision,
            write_pp=settings.write_pp,
            read_pp=settings.read_pp,
            write_LBTE_solution=settings.write_LBTE_solution,
            compression=settings.hdf5_compression,
        )
    else:
        if log_level:
            print(
                "-" * 11
                + " None of ph-ph interaction calculation was performed. "
                + "-" * 11
            )

    _finalize_phono3py(ph3py, confs_dict, log_level, filename=output_yaml_filename)
